from __future__ import division
from scipy import signal
from scipy.signal import filtfilt


def lowpass(s, f, order=2, fs=1000.0, use_filtfilt=False):
    '''
    @brief: for a given signal s rejects (attenuates) the frequencies higher
    then the cuttof frequency f and passes the frequencies lower than that
    value by applying a Butterworth digital filter

    @params:

    s: array-like
    signal

    f: int
    the cutoff frequency

    order: int
    Butterworth filter order

    fs: float
    sampling frequency

    @return:

    signal: array-like
    filtered signal

    '''
    b, a = signal.butter(order, f / (fs/2))

    if use_filtfilt:
        return filtfilt(b, a, s)

    return signal.lfilter(b, a, s)


def highpass(s, f, order=2, fs=1000.0, use_filtfilt=False):
    '''
    @brief: for a given signal s rejects (attenuates) the frequencies lower
    then the cuttof frequency f and passes the frequencies higher than that
    value by applying a Butterworth digital filter

    @params:

    s: array-like
    signal

    f: int
    the cutoff frequency

    order: int
    Butterworth filter order

    fs: float
    sampling frequency

    @return:

    signal: array-like
    filtered signal

    '''

    b, a = signal.butter(order, f * 2 / (fs/2), btype='highpass')
    if use_filtfilt:
        return filtfilt(b, a, s)

    return signal.lfilter(b, a, s)


def bandstop(s, f1, f2, order=2, fs=1000.0, use_filtfilt=False):
    '''
    @brief: for a given signal s rejects (attenuates) the frequencies within a
    certain range (between f1 and f2) and passes the frequencies outside that
    range by applying a Butterworth digital filter

    @params:

    s: array-like
    signal

    f1: int
    the lower cutoff frequency

    f2: int
    the upper cutoff frequency

    order: int
    Butterworth filter order

    fs: float
    sampling frequency

    @return:

    signal: array-like
    filtered signal

    '''
    b, a = signal.butter(order, [f1 * 2 / fs, f2 * 2 / fs], btype='bandstop')
    if use_filtfilt:
        return filtfilt(b, a, s)
    return signal.lfilter(b, a, s)


def bandpass(s, f1, f2, order=2, fs=1000.0, use_filtfilt=False):
    '''
    @brief: for a given signal s passes the frequencies within a certain range
    (between f1 and f2) and rejects (attenuates) the frequencies outside that
    range by applying a Butterworth digital filter

    @params:

    s: array-like
    signal

    f1: int
    the lower cutoff frequency

    f2: int
    the upper cutoff frequency

    order: int
    Butterworth filter order

    fs: float
    sampling frequency

    @return:

    signal: array-like
    filtered signal

    '''
    b, a = signal.butter(order, [f1 * 2 / fs, f2 * 2 / fs], btype='bandpass')

    if use_filtfilt:
        return filtfilt(b, a, s)

    return signal.lfilter(b, a, s)



"""
OpenSignalsTools module intended to give the user the possibility of programmatically generate a
Notebook.

"""

import json
import os
import re
import shutil
import warnings
import importlib.util

import nbformat as nb
from .notebook_files.MainFiles.group_by_difficulty import STAR_TABLE_HEADER
from .notebook_files.MainFiles.group_by_tag import TAG_TABLE_HEADER

from .notebook_files.MainFiles.signal_samples import IMPORT_CODE, PLOTS_SIGNAL_SAMPLES

from .notebook_files.cell_content_strings import DESCRIPTION_GROUP_BY, DESCRIPTION_SIGNAL_SAMPLES, \
                                 HEADER_ALL_CATEGORIES, HEADER_MAIN_FILES, \
                                 DESCRIPTION_CATEGORY, HEADER_TAGS, SEPARATOR, AUX_CODE_MESSAGE, \
                                 JS_CODE_AUTO_PLAY, CSS_STYLE_CODE, FOOTER, HEADER, MD_EXAMPLES, \
                                 CODE_EXAMPLES, OPEN_IMAGE

SIGNAL_TYPE_LIST = ["emg", "ecg"]

NOTEBOOK_KEYS = {"Load": 1, "Record": 2, "Visualise": 3, "Pre-Process": 4, "Detect": 5,
                 "Extract": 6, "Train_and_Classify": 7, "Understand": 8, "Evaluate": 12,
                 "MainFiles": 0}

# ==================================================================================================
# ======================================= notebook Class ===========================================
# ==================================================================================================


class notebook:
    def __init__(self, notebook_type=None, notebook_title="Notebook Title", tags="tags",
                 difficulty_stars=1, notebook_description="Notebook Description",
                 dict_by_difficulty=None, dict_by_tag=None, notebook_file=None):
        """
        Class constructor that generates a new Notebook template, taking into account the
        specified 'notebook_type'.

        ----------
        Parameters
        ----------
        notebook_type : str
            Notebook type: - "Main_Files_Signal_Samples"
                           - "Main_Files_By_Category"
                           - "Main_Files_By_Difficulty"
                           - "Main_Files_By_Tag"
                           - "Main_Files_By_Signal_Type"
                           - "Acquire"
                           - "Open"
                           - "Visualise"
                           - "Process"
                           - "Detect"
                           - "Extract"
                           - "Decide"
                           - "Explain"

        notebook_title : None or str
            The Notebook title should only be defined when 'notebook_type' is:
           - "Acquire"
           - "Open"
           - "Visualise"
           - "Process"
           - "Detect"
           - "Extract"
           - "Decide"
           - "Explain"

        tags : str
            Sequence of tags that characterize the Notebook.

        difficulty_stars : int
            This input defines the difficulty level of the Notebook instructions.

        notebook_description : str
            An introductory text to present the Notebook and involve the reader.

        dict_by_difficulty : dict
            Global Dictionary that groups Notebooks names/files by difficulty level.

        dict_by_tag : dict
            Dictionary where each key is a tag and the respective value will be a list containing the
            Notebooks (title and filename) that include this tag.

        notebook_file : str
            Notebook filename.
        """

        # ========================= Initialisation of Object variables =============================
        self.difficulty_stars = difficulty_stars
        self.tags = tags
        self.notebook_title = notebook_title

        # =============================== Creation of a Notebook ===================================
        self.notebook = nb.v4.new_notebook()

        if "Main" in notebook_type:
            warnings.warn("The arguments 'tags' and 'difficulty_stars' does not have effect for "
                          "the " + notebook_type + " notebook type !")

        if notebook_type in ["Load", "Record", "Visualise", "Pre-Process", "Detect",
                             "Extract", "Train_and_Classify", "Understand", "Evaluate"]:
            self.notebook_type = notebook_type
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_notebook_header(self.notebook, notebook_type, notebook_title, tags,
                                      difficulty_stars, notebook_description)

        elif notebook_type == "Main_Files_By_Category":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_notebooks_by_category(self.notebook, dict_by_tag)
            if os.path.exists("../biosignalsnotebooks_notebooks"):
                _generate_github_readme(self.notebook, dict_by_tag)

        elif notebook_type == "Main_Files_Signal_Samples":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_SIGNAL_SAMPLES)
            _generate_signal_samples_body(self.notebook)

        elif notebook_type == "Main_Files_By_Difficulty":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            #opensignals_hierarchy()
            _generate_notebook_by_difficulty_body(self.notebook, dict_by_difficulty)

        elif notebook_type == "Main_Files_By_Tag":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            # opensignals_hierarchy()
            _generate_notebook_by_tag_body(self.notebook, dict_by_tag)

        elif notebook_type == "Main_Files_By_Signal_Type":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            # opensignals_hierarchy()
            _generate_notebook_by_signal_type_body(self.notebook, dict_by_tag)

    def write_to_file(self, path, filename, footer=True, group_by=False):
        """
        Class method responsible for generating a file containing the notebook object data.

        ----------
        Parameters
        ----------
        path : str
            OpenSignalsTools Root folder path (where the notebook will be stored).

        filename : str
            Defines the name of the notebook file.

        footer : bool
            Flag that defines when the footer needs to be included in the Notebook.

        group_by : bool
            If True then the group by pages that compose the biosignalsnotebooks project will be
            automatically generated (grouping Notebooks by difficulty, tags ...)
        """

        # =============================== Storage of Filename ======================================
        self.filename = filename

        # ======================== Inclusion of Footer in the Notebook =============================
        if footer is True:
            _generate_footer(self.notebook, self.notebook_type)

        # ========== Code segment for application of the OpenSignalsTools CSS style ===========
        self.notebook["cells"].append(nb.v4.new_markdown_cell(AUX_CODE_MESSAGE,
                                                              **{"metadata":
                                                                     {"tags": ["hide_mark",
                                                                               "aux"]}}))
        self.notebook["cells"].append(nb.v4.new_code_cell(CSS_STYLE_CODE,
                                                          **{"metadata":
                                                                 {"tags": ["hide_both"]}}))
        self.notebook["cells"].append(nb.v4.new_code_cell(JS_CODE_AUTO_PLAY,
                                                          **{"metadata":
                                                                 {"tags": ["hide_both"]}}))

        full_path = path + "/Categories/" + self.notebook_type + "/" + filename + ".ipynb"

        # Check if file exists.
        if not os.path.exists(full_path) or self.notebook_type == "MainFiles":
            nb.write(self.notebook, full_path)
        else:
            raise RuntimeError("The specified filename already exists. The Notebook creation was interrupted !")

        # ========================== Run Notebook Code Instructions ================================
        os.system("jupyter nbconvert --execute --inplace --ExecutePreprocessor.timeout=-1 " +
                  full_path)
        os.system("jupyter trust " + full_path)

        # ==== Automatic generation of pages that group the Notebooks by different criteria ========
        if group_by is True:
            _generate_group_by_pages(path)

    def add_markdown_cell(self, content, tags=None):
        """
        Class method responsible for adding a markdown cell with content 'content' to the
        Notebook object.

        ----------
        Parameters
        ----------
        content : str
            Text/HTML code/... to include in the markdown cell (triple quote for multiline text).

        tags : list
            A list of tags to include in the markdown cell metadata.
        """
        self.notebook["cells"].append(nb.v4.new_markdown_cell(content, **{"metadata":
                                                                          {"tags": tags}}))

    def add_code_cell(self, content, tags=None):
        """
        Class method responsible for adding a code cell with content 'content' to the
        Notebook object.

        ----------
        Parameters
        ----------
        content : str
            Code in a string format to include in the cell (triple quote for multiline
            text).

        tags : list
            A list of tags to include in the code cell metadata.
        """
        self.notebook["cells"].append(nb.v4.new_code_cell(content, **{"metadata":
                                                                      {"tags": tags}}))

# ==================================================================================================
# ========================== Generate OpenSignalTools File Hierarchy ===============================
# ==================================================================================================

def opensignals_hierarchy(root=None, update=False, clone=False):
    """
    Function that generates the OpenSignalsTools Notebooks File Hierarchy programatically.

    ----------
    Parameters
    ----------
    root : None or str
        The file path where the OpenSignalsTools Environment will be stored.

    update : bool
        If True the old files will be replaced by the new ones.

    clone : bool
        If True then all the available Notebooks will be stored in the users computer.
        If False only the folder hierarchy of OpenSignalsTools will be generated, giving to the
        user a blank template for creating his own Notebook Environment.

    Returns
    -------
    out : str
        The root file path of OpenSignalsTools Environment is returned.
    """

    if root is None:
        root = os.getcwd()

    categories = list(NOTEBOOK_KEYS.keys())

    # ============================ Creation of the main directory ==================================
    current_dir = root + "/biosignalsnotebooks_environment"
    if not os.path.isdir(current_dir):
        os.makedirs(current_dir)

    # ================== Copy of 'images' 'styles' and 'signal_samples' folders ====================
    package_path = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/")
    for var in ["images", "styles", "signal_samples"]:
        if not os.path.isdir(root + "/biosignalsnotebooks_environment/" + var):
            src = (package_path + "notebook_files/osf_files/" + var).replace("\\", "/")
            destination = current_dir + "/" + var
            shutil.copytree(src, destination)
        elif update is True:
            shutil.rmtree(root + "/biosignalsnotebooks_environment/" + var)
            src = (package_path + "notebook_files/osf_files/" + var).replace("\\", "/")
            destination = current_dir + "/" + var
            shutil.copytree(src, destination)

    # =========================== Generation of 'Categories' folder ================================
    current_dir = root + "/biosignalsnotebooks_environment/Categories"
    if not os.path.isdir(current_dir):
        os.makedirs(current_dir)

    for category in categories:
        if not os.path.isdir(current_dir + "/" + category):
            os.makedirs(current_dir + "/" + category)
            if category == "MainFiles":
                # Copy aux folder.
                src = "../biosignalsnotebooks_notebooks/Categories/MainFiles/aux_files"
                shutil.copytree(src, current_dir + "/" + category + "/aux_files")

    # if clone is True:
    #     # Fill each folder inside "Categories" directory with the respective notebooks.
    #     # Each notebook will be created by a specific function.
    #     dir_path = root + "/notebook_code"
    #     list_of_code_dirs = os.listdir(dir_path)
    #     for folder in list_of_code_dirs:
    #         folder_path = root + "/notebook_code/" + folder
    #         if folder != "MainFiles" and folder != "__pycache__":
    #             list_of_code_files = os.listdir(folder_path)
    #             for file in list_of_code_files:
    #                 if file != "__pycache__":
    #                     spec = importlib.util.spec_from_file_location(file, folder_path +
    #                                                                   "/" + file)
    #                     foo = importlib.util.module_from_spec(spec)
    #                     spec.loader.exec_module(foo)
    #                     foo.run(root + "/biosignalsnotebooks_environment")
    #
    #     # Generation of biosignalsnotebooks environment main files.
    #     main_page = notebook("Main_Files_By_Category")
    #     main_page.write_to_file(root + "/biosignalsnotebooks_environment", "biosignalsnotebooks",
    #                             footer=False)
    #
    #     by_difficulty = notebook("Main_Files_By_Difficulty", "Notebooks Grouped by Difficulty", notebook_description=DESCRIPTION_GROUP_BY)
    #     by_difficulty.write_to_file(root + "/biosignalsnotebooks_environment", "by_diff",
    #                                 footer=False)
    #
    #     by_tags = notebook("Main_Files_By_Tag", "Notebooks Grouped by Tag Values",
    #                        notebook_description=DESCRIPTION_GROUP_BY)
    #     by_tags.write_to_file(root + "/biosignalsnotebooks_environment", "by_tag",
    #                           footer=False)
    #
    #     by_signal_type = notebook("Main_Files_By_Signal_Type", "Notebooks Grouped by Signal Type",
    #                               notebook_description=DESCRIPTION_GROUP_BY)
    #     by_signal_type.write_to_file(root + "/biosignalsnotebooks_environment",
    #                                  "by_signal_type", footer=False)
    #
    #     signal_samples = notebook("Main_Files_Signal_Samples", "Signal Samples Library",
    #                               notebook_description=DESCRIPTION_SIGNAL_SAMPLES)
    #     signal_samples.write_to_file(root + "/biosignalsnotebooks_environment",
    #                                  "signal_samples", footer=False)

    return root + "/biosignalsnotebooks_environment"

# ==================================================================================================
# ========================= Auxiliary Functions for Notebook Generation ============================
# ==================================================================================================


def _generate_notebook_header(notebook_object, notebook_type, notebook_title="Notebook Title",
                              tags="tags", difficulty_stars=1,
                              notebook_description="Notebook Description"):
    """
    Internal function that is used for generation of the generic notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Decide"
                       - "Explain"

    notebook_title : None or str
        The Notebook title should only be defined when 'notebook_type' is:
       - "Acquire"
       - "Open"
       - "Visualise"
       - "Process"
       - "Detect"
       - "Extract"
       - "Decide"
       - "Explain"

    tags : str
        Sequence of tags that characterize the Notebook.

    difficulty_stars : int
        This input defines the difficulty level of the Notebook instructions.

    notebook_description : str
        An introductory text to present the Notebook and involve the reader.

    """

    # ============================= Creation of Header ====================================
    header_temp = HEADER_ALL_CATEGORIES.replace("header_image_color_i", "header_image_color_" +
                                                str(NOTEBOOK_KEYS[notebook_type]))
    header_temp = header_temp.replace("header_image_i", "header_image_" +
                                      str(NOTEBOOK_KEYS[notebook_type]))
    header_temp = header_temp.replace("Notebook Title", notebook_title)
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_temp, **{"metadata": {"tags": ["intro_info_title"]}}))

    # =============== Inclusion of the div with "Difficulty" and "Tags" ===================
    tags_and_diff = HEADER_TAGS.replace('<td class="shield_right" id="tags">tags</td>',
                                        '<td class="shield_right" id="tags">' + "&#9729;".join(tags)
                                        + '</td>')
    for star in range(1, 6):
        if star <= difficulty_stars:
            tags_and_diff = tags_and_diff.replace("fa fa-star " + str(star), "fa fa-star "
                                                                             "checked")
        else:
            tags_and_diff = tags_and_diff.replace("fa fa-star " + str(star), "fa fa-star")

    notebook_object["cells"].append(nb.v4.new_markdown_cell(tags_and_diff,
                                                            **{"metadata": {"tags": ["intro_info_tags"]}}))

    # ================= Insertion of the div reserved to the Notebook Description ==================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(notebook_description,
                                                            **{"metadata":
                                                                {"tags": ["test"]}}))
    notebook_object["cells"].append(nb.v4.new_markdown_cell(SEPARATOR))

    # ======================= Insertion of a blank Markdown and Code cell ==========================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(MD_EXAMPLES))
    notebook_object["cells"].append(nb.v4.new_code_cell(CODE_EXAMPLES))

def _generate_main_files_header(notebook_object, notebook_title="Notebook Title",
                                notebook_description="Notebook Description"):
    """
    Internal function that is used for generation of the 'MainFiles' notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_title : None or str
        Title of the Notebook.

    notebook_description : str
        An introductory text to present the Notebook and involve the reader.

    """

    # =================================== Creation of Header =======================================
    header_temp = HEADER_MAIN_FILES.replace("Notebook Title", notebook_title)
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_temp))

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(notebook_description,
                                                            **{"metadata":
                                                               {"tags": ["test"]}}))
    # notebook_object["cells"].append(nb.v4.new_markdown_cell(SEPARATOR))


def _generate_footer(notebook_object, notebook_type):
    """
    Internal function that is used for generation of the notebooks footer.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the footer will be created.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Decide"
                       - "Explain"

    """

    footer_aux = FOOTER
    if "Main_Files" in notebook_type:
        footer_aux = footer_aux.replace("../MainFiles/", "")

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(footer_aux,
                                                            **{"metadata":
                                                               {"tags": ["footer"]}}))


def _generate_header(notebook_object, notebook_type, notebook_file):
    """
    Internal function that is used for generation of the notebooks footer.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_file : str
        Notebook filename.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Decide"
                       - "Explain"

    """

    header_aux = HEADER
    if notebook_file != None:
        header_aux = header_aux.replace("FILENAME", notebook_file.split(".")[0] + ".dwipynb")
        header_aux = header_aux.replace("SOURCE", "https://mybinder.org/v2/gh/biosignalsnotebooks/biosignalsnotebooks/master?filepath=header_footer%2Fbiosignalsnotebooks_environment%2Fcategories%2F" + notebook_type + "%2F" + notebook_file + ".dwipynb")

    if "Main_Files" in notebook_type:
        header_aux = header_aux.replace("../MainFiles/", "")

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_aux,
                                                            **{"metadata":
                                                               {"tags": ["header"]}}))

def _generate_signal_samples_body(notebook_object):
    """
    Internal function that is used for generation of the 'MainFiles' notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    """

    # =========================== Code Cell for Importing Packages =================================
    notebook_object["cells"].append(nb.v4.new_code_cell(IMPORT_CODE,
                                                        **{"metadata":
                                                            {"tags": ["hide_both"]}}))

    # ========= Generation of a table that synthesises the information about each signal ===========
    signal_samples_dir_jupyter = "../../signal_samples"
    signal_samples_dir_project = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/") + "/notebook_files/osf_files/signal_samples"
    list_of_files = os.listdir(signal_samples_dir_project)
    for file in list_of_files:
        if ".json" not in file:
            file_name = file.split(".")[0]
            with open(signal_samples_dir_project + "/" + file_name + '_info.json') as json_data:
                signal_info = json.load(json_data)

            info_table = "<table width='100%'>\n" \
                         "\t<tr>\n" \
                         "\t\t<td colspan='2' class='signal_samples_header'>" + file_name + \
                         "</td>\n" + "</tr>\n"
            info_keys = list(signal_info.keys())
            for key in info_keys:
                info_table += "\t<tr>\n" \
                              "\t\t<td class='signal_samples_info_keys'>" + key + "</td>\n" + \
                              "\t\t<td class='signal_samples_info_values'>" + signal_info[key] + \
                              "</td>\n" + \
                              "\t</tr>\n"

            info_table += "</table>"

            # ============== Insertion of the HTML table inside a markdown cell ===================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(info_table))

            # ========================= Graphical Representation of Signals =======================
            path_str = '"' + (signal_samples_dir_jupyter + "\\").replace("\\", "/") + '"'
            plot_signal_samples_temp = re.sub(re.compile(r'\bsignal_samples_dir\b'),
                                              path_str, PLOTS_SIGNAL_SAMPLES)
            plot_signal_samples_temp = re.sub(re.compile(r'\bfile\b'),
                                              '"' + str(file) + '"',
                                              plot_signal_samples_temp)
            notebook_object["cells"].append(nb.v4.new_code_cell(plot_signal_samples_temp,
                                                                **{"metadata":
                                                                    {"tags": ["hide_in"]}}))


def _generate_notebook_by_difficulty_body(notebook_object, dict_by_difficulty):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    difficulty level.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_difficulty : dict
        Global Dictionary that groups Notebooks names/files by difficulty level.

    """

    difficulty_keys = list(dict_by_difficulty.keys())
    difficulty_keys.sort()
    for difficulty in difficulty_keys:
        markdown_cell = STAR_TABLE_HEADER
        markdown_cell = _set_star_value(markdown_cell, int(difficulty))
        for notebook_file in dict_by_difficulty[str(difficulty)]:
            split_path = notebook_file.replace("\\", "/").split("/")
            notebook_type = split_path[-2]
            notebook_name = split_path[-1].split("&")[0]
            notebook_title = split_path[-1].split("&")[1]
            markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                             str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                             "src='../../images/icons/" + notebook_type.title() +\
                             ".png' width='15%'>\n\t\t</td>"
            markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                             notebook_title + "\n\t\t</td>"
            markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                             "../" + notebook_type.title() + "/" + notebook_name + \
                             "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

        markdown_cell += "</table>"

        # ==================== Insertion of HTML table in a new Notebook cell ======================
        notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _set_star_value(star_code, number_stars):
    """
    Internal function that is used for update the number of active stars (that define notebook
    difficulty level)

    ----------
    Parameters
    ----------
    star_code : str
        String with the HTML code to be changed.

    number_stars : int
        Number of stars that will be active.

    Returns
    -------
    out : str
        It is returned a string with the HTML code after updating the number of active stars.

    """

    for star in range(1, 6):
        if star <= number_stars:
            star_code = star_code.replace("fa fa-star " + str(star), "fa fa-star "
                                                                             "checked")
        else:
            star_code = star_code.replace("fa fa-star " + str(star), "fa fa-star")

    return star_code


def _generate_notebook_by_tag_body(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    tag values.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    tag_keys = list(dict_by_tag.keys())
    tag_keys.sort()
    for tag in tag_keys:
        if tag.lower() not in SIGNAL_TYPE_LIST:
            markdown_cell = TAG_TABLE_HEADER
            markdown_cell = markdown_cell.replace("Tag i", tag)
            for notebook_file in dict_by_tag[tag]:
                split_path = notebook_file.replace("\\", "/").split("/")
                notebook_type = split_path[-2]
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                                 str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                                 "src='../../images/icons/" + notebook_type.title() +\
                                 ".png' width='15%'>\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                                 notebook_title + "\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                                 "../" + notebook_type.title() + "/" + notebook_name + \
                                 "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

            markdown_cell += "</table>"

            # ==================== Insertion of HTML table in a new Notebook cell ======================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _generate_notebooks_by_category(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page "Notebooks by Category".

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    # ============================ Insertion of an opening text ====================================
    markdown_cell = OPEN_IMAGE

    # == Generation of a table that group Notebooks by category the information about each signal ==
    category_list = list(NOTEBOOK_KEYS.keys())
    tag_keys = list(dict_by_tag.keys())

    markdown_cell += """\n<table id="notebook_list" width="100%">
    <tr>
        <td width="20%" class="center_cell group_by_header_grey"> Category </td>
        <td width="60%" class="center_cell group_by_header"></td>
        <td width="20%" class="center_cell"></td>
    </tr>"""

    for i, category in enumerate(category_list):
        if category != "MainFiles":
            if category.lower() in tag_keys:
                if i == 0:
                    first_border = "color1_top"
                else:
                    first_border = ""


                nbr_notebooks = len(dict_by_tag[category.lower()])
                markdown_cell += "\n\t<tr>" \
                                 "\n\t\t<td rowspan='" + str(nbr_notebooks + 1) + "' class='center_cell open_cell_border_" + str(NOTEBOOK_KEYS[category]) + "'><span style='float:center'><img src='../../images/icons/" + category + ".png' class='icon' style='vertical-align:middle'></span> <span style='float:center' class='color" + str(NOTEBOOK_KEYS[category]) + "'>" + category + "</span></td>" \
                                 "\n\t\t<td class='center_cell color" + str(NOTEBOOK_KEYS[category]) + "_cell " + first_border + "'><span style='float:center'>" + category +  "</span></td>" \
                                 "\n\t\t<td class='center_cell gradient_color" + str(NOTEBOOK_KEYS[category]) + "'></td>" \
                                 "\n\t</tr>"

                notebook_list = dict_by_tag[category.lower()]
                for j, notebook_file in enumerate(notebook_list):
                    if j == len(notebook_list) - 1:
                        last_border = "class='border_cell_bottom_white'"
                    else:
                        last_border = ""

                    split_path = notebook_file.replace("\\", "/").split("/")
                    notebook_name = split_path[-1].split("&")[0]
                    notebook_title = split_path[-1].split("&")[1]
                    markdown_cell += "\n\t<tr " + last_border + ">" \
                                     "\n\t\t<td class='center_cell open_cell_light'> <a href='../" + category + "/" + notebook_name + "'>" + notebook_title + "</a> </td>" \
                                     "\n\t\t<td class='center_cell'> <a href='../" + category + "/" + notebook_name + "'><div class='file_icon'></div></a> </td>" \
                                     "\n\t</tr>"

    markdown_cell += "\n</table>"

    # ============================ Insertion of an introductory text ===============================
    markdown_cell += DESCRIPTION_CATEGORY

    # =================== Insertion of the HTML table inside a markdown cell =======================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _generate_github_readme(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the GitHub and PyPI README file.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    # ==================== List of urls that contain each Category icon ============================
    icons = {"Detect": "https://i.ibb.co/rymrvFL/Detect.png",
             "Evaluate": "https://i.ibb.co/yfwcy2M/Evaluate.png",
             "Extract": "https://i.ibb.co/tchq7Cc/Extract.png",
             "Load": "https://i.ibb.co/YPbCnzD/Load.png",
             "Pre-Process": "https://i.ibb.co/1rKWccX/Pre-Process.png",
             "Record": "https://i.ibb.co/d2jZH1s/Record.png",
             "Train_and_Classify": "https://i.ibb.co/CQ4cyGb/Train-and-Classify.png",
             "Understand": "https://i.ibb.co/MnhRRQT/Understand.png",
             "Visualise": "https://i.ibb.co/wh4HKzf/Visualise.png"}

    # =========================== biosignalsnotebooks website ======================================
    biosignalsnotebooks_web = "http://www.biosignalsplux.com/notebooks/Categories/"

    # == Generation of a table that group Notebooks by category the information about each signal ==
    category_list = list(NOTEBOOK_KEYS.keys())
    tag_keys = list(dict_by_tag.keys())

    markdown_cell = """<table width="100%">
    <tr>
        <td width="20%" align="center"><strong> Category <strong></td>
        <td width="80%"></td>
    </tr>"""

    for i, category in enumerate(category_list):
        if category != "MainFiles":
            if category.lower() in tag_keys:
                nbr_notebooks = len(dict_by_tag[category.lower()])
                notebook_list = dict_by_tag[category.lower()]
                split_path = notebook_list[0].replace("\\", "/").split("/")
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\n\t<tr>" \
                                 "\n\t\t<td rowspan='" + str(nbr_notebooks) + "'><p align='center'><img src='" + icons[category] + "' width='50%' align='center'></p></td>" \
                                 "\n\t\t<td align='center'> <a href='" + biosignalsnotebooks_web + category + "/" + notebook_name.replace(".ipynb", "_rev.php") + "' target='_blank'>" + notebook_title + "</a> </td>" \
                                 "\n\t</tr>"

                for j, notebook_file in enumerate(notebook_list[1:]):
                    split_path = notebook_file.replace("\\", "/").split("/")
                    notebook_name = split_path[-1].split("&")[0]
                    notebook_title = split_path[-1].split("&")[1]
                    markdown_cell += "\n\t<tr>" \
                                     "\n\t\t<td align='center'> <a href='" + biosignalsnotebooks_web + category + "/" + notebook_name.replace(".ipynb", "_rev.php") + "'>" + notebook_title + "</a> </td>" \
                                     "\n\t</tr>"

    markdown_cell += "\n</table>"

    # ============================ Generation of README files ======================================
    # [Open template]
    template_path = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/") + "/notebook_files/github/README_TEMPLATE.md"
    with open(template_path, 'r') as readme:
        readme_str = readme.read()

    readme_str = readme_str.replace("LIST_OF_NOTEBOOKS", markdown_cell)

    # [Storage of the updated files]
    for path in ["../biosignalsnotebooks/README_BSN.md", "../README.md"]:
        with open(path, 'w') as readme_out:
            readme_out.write(readme_str)
        readme_out.close()


def _generate_notebook_by_signal_type_body(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    signal type where they are applicable.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.
    """

    tag_keys = list(dict_by_tag.keys())
    tag_keys.sort()
    for tag in tag_keys:
        if tag.lower() in SIGNAL_TYPE_LIST:
            markdown_cell = TAG_TABLE_HEADER
            markdown_cell = markdown_cell.replace("Tag i", tag.upper())
            for notebook_file in dict_by_tag[tag]:
                split_path = notebook_file.replace("\\", "/").split("/")
                notebook_type = split_path[-2]
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                                 str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                                                                     "src='../../images/icons/" + notebook_type.title() + \
                                 ".png' width='15%'>\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                                 notebook_title + "\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                                 "../" + notebook_type.title() + "/" + notebook_name + \
                                 "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

            markdown_cell += "</table>"

            # ==================== Insertion of HTML table in a new Notebook cell ======================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))

def _generate_group_by_pages(root):
    dict_group_by_diff, dict_group_by_tag = _search_in_notebooks(root)
    file_path = root

    # Generation of biosignalsnotebooks environment main files.
    filename = "biosignalsnotebooks"
    main_page = notebook("Main_Files_By_Category", dict_by_difficulty=dict_group_by_diff,
                         dict_by_tag=dict_group_by_tag, notebook_file=filename)
    main_page.write_to_file(file_path, filename)
    filename = "by_diff"
    by_difficulty = notebook("Main_Files_By_Difficulty", "Notebooks Grouped by Difficulty",
                             dict_by_difficulty=dict_group_by_diff,
                             dict_by_tag=dict_group_by_tag,
                             notebook_file=filename)
    by_difficulty.write_to_file(file_path, filename)

    filename = "by_tag"
    by_tags = notebook("Main_Files_By_Tag", "Notebooks Grouped by Tag Values",
                       dict_by_difficulty=dict_group_by_diff, dict_by_tag=dict_group_by_tag,
                       notebook_file=filename)
    by_tags.write_to_file(file_path, filename)

    filename = "by_signal_type"
    by_signal_type = notebook("Main_Files_By_Signal_Type",
                              "Notebooks Grouped by Signal Type",
                              dict_by_difficulty=dict_group_by_diff,
                              dict_by_tag=dict_group_by_tag,
                              notebook_file=filename)
    by_signal_type.write_to_file(file_path, filename)

    filename = "signal_samples"
    signal_samples = notebook("Main_Files_Signal_Samples", "Signal Samples Library",
                              notebook_file=filename)
    signal_samples.write_to_file(file_path, filename)

def _search_in_notebooks(root):
    # Variable definition.
    dict_group_by_diff = {}
    dict_group_by_tag = {}

    categories = os.listdir(root + "/Categories")
    for category in categories:
        current_folder_path = root + "/Categories/" + category
        list_files = os.listdir(current_folder_path)
        for file in list_files:
            current_file_path = current_folder_path + "/" + file
            # Access to all Notebook files (with the extension .ipynb)
            if file.endswith(".ipynb"):
                # Read of the current Notebook.
                notebook = nb.read(current_file_path, nb.NO_CONVERT)

                # Search for "header" and/or "footer".
                header_cell, footer_cell, title, nbr_stars, tags = _get_metadata(notebook, file,
                                                                                 category)

                # Storage of Notebook metadata in global dictionaries.
                if category != "MainFiles":
                    if str(nbr_stars) not in (dict_group_by_diff.keys()):
                        dict_group_by_diff[str(nbr_stars)] = []
                    dict_group_by_diff[str(nbr_stars)].append(current_file_path + "&" + title)

                    for tag in tags:
                        if tag not in (dict_group_by_tag.keys()):
                            dict_group_by_tag[str(tag)] = []
                        dict_group_by_tag[str(tag)].append(current_file_path + "&" + title)

    return dict_group_by_diff, dict_group_by_tag


def _get_metadata(notebook, filename, category):
    # Variable initialisation.
    header_cell = None
    footer_cell = None
    title = None
    nbr_stars = None
    tags = None
    list_cells = notebook["cells"]

    # Search for "header" and "footer" tags and collect Notebook "difficulty" and "tags".
    for cell_nbr, cell in enumerate(list_cells):
        if "tags" in list(cell["metadata"].keys()):
            # [Header and Footer identification]
            if "header" in cell["metadata"]["tags"] and header_cell is None:
                header_cell = cell_nbr
            elif "footer" in cell["metadata"]["tags"] and footer_cell is None:
                footer_cell = cell_nbr
            elif "aux" in cell["metadata"]["tags"] and footer_cell is None:
                footer_cell = cell_nbr
            elif ("header" in cell["metadata"]["tags"] and header_cell is not None) or \
                    ("footer" in cell["metadata"]["tags"] and footer_cell is not None):
                raise RuntimeError("Duplicated 'header' or 'cell' tags inside the Notebook " +
                               filename + " !")

            # [Title, Difficulty and Tags]
            if category != "MainFiles":
                # [Title]
                if "intro_info_title" in cell["metadata"]["tags"]:
                    cell_content = cell["source"]

                    # Notebook Title.
                    title = cell_content.split('<td class="header_text">')[1].split('</td>')[0]

                # [Tags]
                if "intro_info_tags" in cell["metadata"]["tags"]:
                    cell_content = cell["source"]

                    # Notebook tag list.
                    tags = cell_content.split('<td class="shield_right" id="tags">')[1].split('</td>')[0].split("&#9729;")

                    # Difficulty level.
                    nbr_stars = cell_content.count("checked")

    return header_cell, footer_cell, title, nbr_stars, tags

# 29/11/2018  17h18m :)



"""
    Private module uniquely dedicated to clone the last version of bsnb_files, from the
    biosignalsnotebooks original folder.

"""

import os
import shutil

def run_update():
    current_dir = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/")

    # =============== Update of "images", "signal_samples" and "styles" directories ================
    aux_dir = current_dir + "osf_files/"
    for folder in ["images", "signal_samples", "styles"]:
        # Removal of oldest directory versions.
        if os.path.isdir(aux_dir + folder):
            shutil.rmtree(aux_dir + folder)

        # Clone the updated folder.
        src = "../../../biosignalsnotebooks_notebooks/" + folder
        if os.path.isdir(src):
            shutil.copytree(src, aux_dir + folder)
        else:
            raise RuntimeError("The current module has a private state. It should only be used by "
                               "the package developers!")



run_update()



# THIS FILE IS GENERATED FROM pyedflib SETUP.PY
short_version = '0.1.14'
version = '0.1.14'
full_version = '0.1.14'
git_revision = 'Unknown'
release = True

if not release:
    version = full_version



"""
biosignalsnotebooks module responsible for loading data from the available signal samples (used on
biosignalsnotebooks Notebooks).

Available Functions
-------------------
[Public]

load_signal
    Function that returns a dictionary with the data contained inside a signal sample file
    (stored in the biosignalsnotebooks signal samples directory).

[Private]

_generate_download_google_link
    Google Drive sharable links do not allow a direct download with wget function, so, this private
    function manipulates the input link to ensure the desired download.

Observations/Comments
---------------------
None

/\
"""

# =================================================================================================
# ====================================== Import Statements ========================================
# =================================================================================================
import os
from .load import load
from .aux_functions import _generate_download_google_link

SIGNAL_PATH = (os.path.abspath(__file__).split(os.path.basename(__file__))[0] + \
              "notebook_files\\osf_files\\signal_samples\\").replace("\\", "/")
FILE_EXTENSION = ".h5"


def load_signal(signal_handler, get_header=False):
    """
    Function that returns a dictionary with the data contained inside 'signal_name' file (stored in
    the biosignalsnotebooks signal samples directory.

    ----------
    Parameters
    ----------
    signal_name : file name or url
        Name that identifies the signal sample to be loaded or a url.

        Possible values:
        [ecg_4000_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:12.4
        Sample Rate                4000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_5_min]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           05:00.0
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_sample]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:11.9
        Sample Rate                 200 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_20_sec_10_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:20.0
        Sample Rate                  10 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [ecg_20_sec_100_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:19.7
        Sample Rate                 100 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [ecg_20_sec_1000_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:20.4
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [emg_bursts]
        =================   ==============
        Signal Type                    EMG
        Muscle              Biceps Brachii
        Acquisition Time           00:28.5
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions             Cyclic
                               Contraction
        =================   ==============

        [emg_fatigue]
        =================   ==============
        Signal Type                    EMG
        Muscle              Biceps Brachii
        Acquisition Time           02:06.9
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions          Cyclic Flexion
                            and Extension
                            for fatigue
                            induction
        =================   ==============

        [temp_res_8_16]
        =================   ==============
        Signal Type            Temperature
        Acquisition Time           03:53.1
        Sample Rate                1000 Hz
        Number of Channels               2
        Resolutions          8 and 16 bits
        Conditions           Temperature
                             increase and
                             decrease
        =================   ==============

        [bvp_sample]
        =================   ==============
        Signal Type                    BVP
        Acquisition Time           00:27.3
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

    get_header : boolean
        If true the file header will be returned as one of the function outputs.

    Returns
    -------
    out : dict
        A dictionary with the data stored inside the file specified in the input 'signal_name'.

    header : dict
        Metadata of the acquisition file (includes sampling rate, resolution, used device...
    """

    available_signals = ["ecg_4000_Hz", "ecg_5_min", "ecg_sample", "ecg_20_sec_10_Hz",
                         "ecg_20_sec_100_Hz", "ecg_20_sec_1000_Hz", "emg_bursts", "emg_fatigue",
                         "temp_res_8_16", "bvp_sample"]

    # Check if signal_handler is a url.
    # [Statements to be executed if signal_handler is a url]
    if any(mark in signal_handler for mark in ["http://", "https://", "www.", ".pt", ".com", ".org",
                                               ".net"]):
        # Check if it is a Google Drive sharable link.
        if "drive.google" in signal_handler:
            signal_handler = _generate_download_google_link(signal_handler)

        # Load file.
        out, header = load(signal_handler, remote=True, get_header=True)

    # [Statements to be executed if signal_handler is an identifier of the signal]
    else:
        if signal_handler in available_signals:
            out, header = load(SIGNAL_PATH + signal_handler + FILE_EXTENSION, get_header=True)
        else:
            raise RuntimeError("The signal name defined as input does not correspond to any of the "
                           "signal samples contained in the package.")

    if get_header is True:
        return out, header
    else:
        return out

# 11/10/2018 16h45m :)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from pylab import rc, close, figure, axes, subplot, plot, axis, show, grid, savefig, text
from numpy import arange
#added import - David Belo
from matplotlib import pyplot as plot
import pandas
import tkinter.filedialog as tkFileDialog

def load_data_dialog(path):
    #root = Tkinter.Tk()
    #root.withdraw()
    
    # Make it almost invisible - no decorations, 0 size, top left corner.
    #root.overrideredirect(True)
    #root.geometry('0x0+0+0')
    
    # Show window again and lift it to top so it can get focus,
    # otherwise dialogs will end up behind the terminal.
    #root.deiconify()
    #root.lift()
    #root.focus_force()
    
    #filename = tkFileDialog.askopenfile(parent=root) # Or some other dialog
    filename = tkFileDialog.askopenfile() # Or some other dialog
    
    # Get rid of the top-level instance once to make it actually invisible.
    #root.destroy()
    return pandas.read_csv(filename, sep=' ', header=None)
    

##########
#Initial configurations
def pylabconfig():

    rc('lines', linewidth=2, color='k')
    #rc('lines', linewidth=1, color='k')

    rc('font', **{'family': 'serif', 'serif': ['Palatino']})
    rc('font', style='italic', size=10)

    rc('text', color='grey')

    #       rc('text', usetex=True)

    rc('text', usetex=False)

    rc('figure', figsize=(8, 5), dpi=80)
    rc('axes', grid=True, edgecolor='grey', labelsize=10,)
    rc('grid', color='grey')
    rc('xtick', color='grey', labelsize=10)
    rc('ytick', color='grey', labelsize=10)

    close('all')



    
def plotwithhist(t, s, bins=50):
    from matplotlib.ticker import NullFormatter

    nullfmt = NullFormatter()
    figure()
    ax2 = axes([0.125 + 0.5, 0.1, 0.2, 0.8])
    ax1 = axes([0.125, 0.1, 0.5, 0.8])

    ax1.plot(t, s)
    ax1.set_xticks(ax1.get_xticks()[:-1])

    ax2.hist(s, bins, normed=True, facecolor='white',
             orientation='horizontal', lw=2)

    ax2.axis([0, 1, ax1.axis()[2], ax1.axis()[3]])

    ax2.yaxis.set_major_formatter(nullfmt)
    ax2.set_xticks([0, 0.5, 1])

    return ax1, ax2


    ###########

def plotwithstats(t, s):
    
    from matplotlib.ticker import NullFormatter

    nullfmt = NullFormatter()
    figure()
    ax2 = axes([0.125 + 0.5, 0.1, 0.2, 0.8])

    ax1 = axes([0.125, 0.1, 0.5, 0.8])

    ax1.plot(t, s)
    ax1.set_xticks(ax1.get_xticks()[:-1])

    meanv = s.mean()
    mi = s.min()
    mx = s.max()
    sd = s.std()

    ax2.bar(-0.5, mx - mi, 1, mi, lw=2, color='#f0f0f0')
    ax2.bar(-0.5, sd * 2, 1, meanv - sd, lw=2, color='#c0c0c0')
    ax2.bar(-0.5, 0.2, 1, meanv - 0.1, lw=2, color='#b0b0b0')
    ax2.axis([-1, 1, ax1.axis()[2], ax1.axis()[3]])

    ax2.yaxis.set_major_formatter(nullfmt)
    ax2.set_xticks([])

    return ax1, ax2


def multilineplot(signal, linesize=250, events=None, title = '', dir = '', step=1):
    from pylab import rc
    rc('axes', labelcolor='#a1a1a1', edgecolor='#a1a1a1', labelsize='xx-small')
    rc('xtick',color='#a1a1a1', labelsize='xx-small')

    
    grid('off')
    nplots=len(signal)//linesize + 1
    ma_x = max(signal)
    mi_x = min(signal)
    f=figure(figsize=(20, 1.5*nplots), dpi=80)
    for i in range(nplots):
        ax=subplot(nplots,1,i+1)
        
        start = i*len(signal)/nplots
        end = (i+1)*len(signal)/nplots
        plot(arange(start,end,step),signal[start:end:step],'k')
        axis((start,end,mi_x,ma_x))
        ax.set_yticks([])
        ax.set_xticks(ax.get_xticks()[1:-1])
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        
        #ax.spines['bottom'].set_visible(False)
        #ax.spines['left'].set_bounds(mi_x,ma_x)
        ax.xaxis.set_ticks_position('bottom')
        if events != None:
            e = events[(events >= start) & (events < end)]

            if len(e)>0:
                plot.vlines(e,mi_x,ma_x-(ma_x-mi_x)/4.*3., lw=2)

        if title != None:
            text(start,ma_x,title)
        grid('off')
    f.tight_layout()
    savefig(dir+title+'.pdf')
    close()
 



from pylab import axis, draw, close, gcf, gca


def zoom(event):
    ax = gca()
    cur_xlim = ax.get_xlim()
    cur_ylim = ax.get_ylim()

    # edit the scale if needed
    base_scale = 1.1

    xdata = event.xdata     # get event x location
    ydata = event.ydata     # get event y location

    # performs a prior check in order to not exceed figure margins
    if xdata != None and ydata != None:
        if event.button == 'up':
            # deal with zoom in
            scale_factor = 1 / base_scale
        elif event.button == 'down':
            # deal with zoom out
            scale_factor = base_scale
        else:
            # deal with something that should never happen
            scale_factor = 1
            print(event.button)

        new_width = (cur_xlim[1] - cur_xlim[0]) * scale_factor
        new_height = (cur_ylim[1] - cur_ylim[0]) * scale_factor

        relx = (cur_xlim[1] - xdata)/(cur_xlim[1] - cur_xlim[0])
        rely = (cur_ylim[1] - ydata)/(cur_ylim[1] - cur_ylim[0])

        ax.set_xlim([xdata - new_width * (1-relx), xdata + new_width * (relx)])
        ax.set_ylim([ydata - new_height * (1-rely), ydata + new_height * (rely)])
        ax.figure.canvas.draw()

    return zoom


def on_key_press(event):

    # keyboard zoom-in
    if event.key == '+':
        a = axis()
        w = a[1] - a[0]
        axis([a[0] + w * .2, a[1] - w * .2, a[2], a[3]])
        draw()

    # keyboard zoom-out
    if event.key in ['-', '\'']:
        a = axis()
        w = a[1] - a[0]
        axis([a[0] - w / 3.0, a[1] + w / 3.0, a[2], a[3]])
        draw()

    # right displacement
    if event.key in ['.', 'right']:
        a = axis()
        w = a[1] - a[0]
        axis([a[0] + w * .2, a[1] + w * .2, a[2], a[3]])
        draw()

    # left displacement
    if event.key in [',', 'left']:
        a = axis()
        w = a[1] - a[0]
        axis([a[0] - w * .2, a[1] - w * .2, a[2], a[3]])
        draw()

    # up displacement
    if event.key == 'up':
        a = axis()
        w = a[3] - a[2]
        axis([a[0], a[1], a[2] + w * .2, a[3] + w * .2])
        draw()

    # down displacement
    if event.key == 'down':
        a = axis()
        w = a[3] - a[2]
        axis([a[0], a[1], a[2] - w * .2, a[3] - w * .2])
        draw()

    # close figure
    if event.key == 'q':
        close()
        # NOTE: We should make the disconnect (mpl_disconect(cid)
        # But since the figure is destroyed we may keep this format
        # if implemented the mpl_connect should use the return cid

    # print('you pressed', event.key, event.xdata, event.ydata)
    # TODO: Reset zoom with an initial default value -> suggest 'r' key


def on_key_release(event):
    # print('you released', event.key, event.xdata, event.ydata)
    pass


def niplot():
    """
    This script extends the native matplolib keyboard bindings.
    This script allows to use the `up`, `down`, `left`, and `right` keys
    to move the visualization window. Zooming can be performed using the `+`
    and `-` keys. Finally, the scroll wheel can be used to zoom under cursor.

    Returns
    -------

    """
    fig = gcf()
    cid = fig.canvas.mpl_connect('key_press_event',  # @UnusedVariable
                                 on_key_press)
    cid = fig.canvas.mpl_connect('key_release_event',  # @UnusedVariable
                                 on_key_release)
    cid = fig.canvas.mpl_connect('scroll_event', zoom)




from numpy import *

def sumvolve(x,window):
    lw=len(window)
    res=zeros(len(x)-lw,'d')
    for i in range(len(x)-lw):
        res[i]=sum(abs(x[i:i+lw]-window))/float(lw)
        #res[i]=sum(abs(x[i:i+lw]*window))
    return res



"""
Includes a set of auxiliary functions that are invoked in other biosignalsnotebooks modules.

Like the module name suggests, these functions have a supporting role and their use is not
restricted to a single module or category of biosignalsnotebooks Jupyter Notebooks.

Considering this generic nature, the great majority of the functions contained inside
"aux_functions" module are classified as private and his inclusion in the package is only intended
for the use inside biosignalsnotebooks project, though they are completely functional for other
applications.

Available Functions
-------------------
[Private]

_is_instance
    Function responsible for checking when the type of 'all' or 'any' list (input) elements is equal
    to a specific data type (input).
_butter_bandpass_filter
    Generation of a Butterworth bandpass filter and application of a signal to it.
_moving_average
    Function dedicated to the application of a moving average filter, intended to smooth the EMG
    signal.
_filter_keywords
    Function for filtering kwargs, ensuring that invalid keywords are not passed as arguments when
    certain function is called.
_inv_key
    It is invoked by _filter_keywords and used for identification when a list of keywords contain
    invalid keywords.


Observations/Comments
---------------------
None

/\
"""

import numpy
import scipy.signal as scisign
from inspect import signature
from warnings import warn
import time as time_package
import os
from bokeh.plotting import output_file


def _is_instance(type_to_check, element, condition="any", deep=False):
    """
    -----
    Brief
    -----
    Function that verifies when "all" or "any" elements of the list "element" have the type
    specified in "type_to_check" input.

    -----------
    Description
    -----------
    In some biosignalsnotebooks functions their implementation is extremely dependent on a specific
    criterion, i.e., 'all' list entries should be of a specific data type.

    In order to ensure this functionality _is_instance function was implemented.

    For example, when plotting data through 'plot' function of 'visualise' module, 'all' entries
    of time axis and data samples lists need to be 'Numeric'.
    In order to this condition be checked _is_instance should be called with the following input
    values:

    _is_instance(Number, [1, 2, 3, True, ...], 'all')

    Sometimes is also relevant to check if at least one of list entries belongs to a data type, for
    cases like this, the argument "condition" should have value equal to "any".

    --------
    Examples
    --------
    >>> _is_instance(Number, [1, 2, 3, True], 'all')
    False
    >>> _is_instance(Number, [1, 1.2, 3, 5], 'all')
    True

    ----------
    Parameters
    ----------
    type_to_check : type element
        Data type (all or any elements of 'element' list must be of the type specified in the
        current input).

    element : list
        List where condition specified in "condition" will be checked.

    condition : str
        String with values "any" or "all" verifying when "any" or "all" element entries have the
        specified type.

    deep : bool
        Flag that identifies when element is in a matrix format and each of his elements should be
        verified iteratively.

    Returns
    -------
    out : boolean
        Returns True when the "condition" is verified for the entries of "element" list.
    """

    out = None

    # Direct check of "condition" in "element".
    if deep is False:
        if condition == "any":
            out = any(isinstance(el, type_to_check) for el in element)
        elif condition == "all":
            out = all(isinstance(el, type_to_check) for el in element)

    # Since "element" is in a matrix format, then it will be necessary to check each dimension.
    else:
        for row in range(0, len(element)):
            for column in range(0, len(element[row])):
                flag = _is_instance(type_to_check, element[column][row], "all", deep=False)
                if flag is False:
                    out = flag
                else:
                    out = True
    return out


def _butter_bandpass_filter(signal, low_cutoff, high_cutoff, sample_rate, order=6):
    """
    -----
    Brief
    -----
    Function with the purpose of applying a digital bandpass filter on signal, in order to attenuate
    the signal content outside the frequency band [low_cutoff, high_cutoff].

    -----------
    Description
    -----------
    In signal processing knowledge can be extracted from different domains, the most common is the
    time-domain, where statistical parameters are taken by analysing the acquired samples of
    a time series.

    Another common domain of signal analysis is the frequency domain, also known as
    "Fourier Domain" due to the fact that for converting data from time to frequency domain it is
    necessary to apply the Fourier Transform.

    The basic principle behind the Fourier Analysis is related with the mathematical decomposition
    of a time-series.

    As demonstrated by Joseph Fourier in his book "Théorie analytique de la chaleur":
    "some functions could be written as an infinite sum of harmonics".

    Each "harmonic" is characterized by a specific frequency, so, after decomposing a time series
    into their harmonics/elementary components, the "frequency content" of the signal is revealed.

    If the frequency content is manipulated, for example, by decreasing the weight of some
    harmonics, after the reconstruction ("sum of harmonics") then the resultant signal in time
    domain will be slightly different comparing with the original one.

    With the current function an input "signal" is applied to a butterworth bandpass filter, in
    order to decrease the weight of elementary components with frequencies smaller than "low_cutoff"
    or bigger than "high_cutoff".

    Filter order can also be specified (the bigger the order more quickly the response of the filter
    will be, near the cutoff frequencies). Higher orders could cause a less stable behavior of the
    system.

    ----------
    Parameters
    ----------
    signal : ndarray
        Array that contains the acquired samples of the signal under analysis.
    low_cutoff : int
        Low cutoff frequency of the bandpass filter.
    high_cutoff : int
        High cutoff frequency of the bandpass filter.
    sample_rate : int
        Sampling rate at which the acquisition took place.
    order : int
        Filter order.

    Returns
    -------
    out : list
        Filtered signal samples.
    """

    order = order
    nyquist_freq = 0.5 * sample_rate

    # Transfer Function.
    b_coeff, a_coeff = scisign.butter(order, [low_cutoff / nyquist_freq,
                                              high_cutoff / nyquist_freq], btype='band')[:2]

    # Application of the signal to the designed filter.
    sign = scisign.lfilter(b_coeff, a_coeff, signal)

    return sign


def _moving_average(data, wind_size=3):
    """
    -----
    Brief
    -----
    Application of a moving average filter for signal smoothing.

    -----------
    Description
    -----------
    In certain situations it will be interesting to simplify a signal, particularly in cases where
    some events with a random nature take place (the random nature of EMG activation periods is
    a good example).

    One possible simplification procedure consists in smoothing the signal in order to obtain
    only an "envelope". With this methodology the analysis is mainly centered on seeing patterns
    in data and excluding noise or rapid events [1].

    The simplification can be achieved by segmenting the time series in multiple windows and
    from each window an average value of all the samples that it contains will be determined
    (dividing the sum of all sample values by the window size).

    A quick and efficient implementation (chosen in biosignalsnotebooks package) of the moving window
    methodology is through a cumulative sum array.

    [1] https://en.wikipedia.org/wiki/Smoothing

    ---------
    Parameters
    ----------
    data : list
        List of signal samples.
    wind_size : int
        Number of samples inside the moving average window (a bigger value implies a smoother
        output signal).

    Returns
    -------
    out : numpy array
        Array that contains the samples of the smoothed signal.
    """

    wind_size = int(wind_size)
    ret = numpy.cumsum(data, dtype=float)
    ret[wind_size:] = ret[wind_size:] - ret[:-wind_size]
    return numpy.concatenate((numpy.zeros(wind_size - 1), ret[wind_size - 1:] / wind_size))


def _filter_keywords(function, kwargs_dict, is_class=False, warn_print=True):
    """
    -----
    Brief
    -----
    Function for filtering kwargs, ensuring that invalid keywords are not passed as arguments.

    -----------
    Description
    -----------
    Many Python functions included in the major packages, include in their definition some
    keyword arguments (kwargs), i.e., a dictionary having as keys non-formal parameters.

    In spite of not being formally defined, only certain keywords are accepted.

    In order to avoid that the biosignalsnotebooks user gets some errors due to an invalid keyword
    argument, when invoking functions such as 'plot' ('visualise' module), _filter_keywords receives
    the keywords entered by the user, checks if these keywords are valid and only returns a
    dictionary with the valid ones.

    ----------
    Parameters
    ----------
    function : function name
        Function where the **kwargs dictionary must be applied.

    kwargs_dict : dict
        Variable keyword arguments.

    Returns
    -------
    kwargs_final : dict
        A filtered dictionary of valid kwargs for using in the specified function.
    """

    # %%%%%%%%%%%%%%%%%%%%%%%%%% List of keywords of our input function %%%%%%%%%%%%%%%%%%%%%%%%%%%
    if is_class is False:
        list_of_keywords = list(signature(function).parameters.keys())
    else:
        list_of_keywords = list(function.__dict__.keys())

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    # %% List of keywords passed in **kwargs that match the valid keywords of the input function %%
    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    available_keywords = set(kwargs_dict.keys())
    valid_keywords = list(available_keywords.intersection(list_of_keywords))

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Final kwargs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    bool_val, inv_keys = _inv_key(list(available_keywords), list_of_keywords)
    if not bool_val:
        str_inval_keys = ""
        for inval_key in inv_keys:
            str_inval_keys += " " + str(inval_key) + ";"
            if warn_print is True:
                warn(RuntimeWarning("At least one of the specified kwargs is not applied "
                                    "in numpy.loadtxt function" + str_inval_keys),
                     stacklevel=2)
                break

    kwargs_final = dict((your_key, kwargs_dict[your_key]) for your_key in valid_keywords)

    return kwargs_final


def _inv_key(list_keys, valid_keys):
    """
    -----
    Brief
    -----
    A sub-function of _filter_keywords function.

    -----------
    Description
    -----------
    Function used for identification when a list of keywords contain invalid keywords not present
    in the valid list.

    ----------
    Parameters
    ----------
    list_keys : list
        List of keywords that must be verified, i.e., all the inputs needs to be inside valid_keys
        in order to a True boolean be returned.

    valid_keys : list
        List of valid keywords.

    Returns
    -------
    out : boolean, list
        Boolean indicating if all the inserted keywords are valid. If true a list with invalid
        keywords will be returned.
    """

    inv_keys = []
    bool_out = True
    for i in list_keys:
        if i not in valid_keys:
            bool_out = False
            inv_keys.append(i)

    return bool_out, inv_keys

def _generate_bokeh_file(file_name):
    """
    -----
    Brief
    -----
    Auxiliary function responsible for the creation of a directory where Bokeh figures will be
    stored.
    The "active" output file for Bokeh will also be updated for the new one.

    -----------
    Description
    -----------
    To ensure that Bokeh plots are correctly observed in the HTML version of the Notebooks, it is
    necessary to embed the plots inside Iframes.

    Taking this into consideration, the source file of the plot is mandatory to use an Iframe, and
    this function ensures the generation of a Bokeh file for each plot, storing it in an adequate
    place.

    ----------
    Parameters
    ----------
    file_name : str
        Name given to the file.

    Returns
    -------
    out : str
        String containing the file name.
    """
    # Creation of our output file instance.
    if file_name is None:
        file_name = "plot_" + time_package.strftime("%Y_%m_%d_%H_%M_%S.html")
    else:
        file_name += ".html"

    if not os.path.exists("generated_plots"):
        os.makedirs("generated_plots")

    output_file(os.getcwd().replace("\\", "/") + "/generated_plots/" + file_name)

    return file_name


def _is_a_url(input_element):
    """
    -----
    Brief
    -----
    Auxiliary function responsible for checking if the input is a string that contains a url.

    -----------
    Description
    -----------
    Some biosignalsnotebooks functions support a remote access to files. In this situation it is
    important to understand if the input is a url, which can be easily achieved by checking if some
    key-markers are present.

    The key-markers that will be searched are "http://", "https://", "www.", ".pt", ".com", ".org",
    ".net".

    ----------
    Parameters
    ----------
    input_element : unknown
        The data structure that will be checked.

    Returns
    -------
    out : bool
        If the input_element is a string and if it contains any key-marker, then True flag will be returned.
    """
    if type(input_element) is str:
        # Check if signal_handler is a url.
        # [Statements to be executed if signal_handler is a url]
        if any(mark in input_element for mark in ["http://", "https://", "www.", ".pt", ".com",
                                                  ".org", ".net"]):
            return True
        else:
            return False
    else:
        return False

def _generate_download_google_link(link):
    """
    -----
    Brief
    -----
    Function that returns a direct download link of a file stored inside a Google Drive
    Repository.

    -----------
    Description
    -----------
    Generally a link from a Google Drive file is only for viewing purposes.

    If the user wants to download the file it can be done with Google Drive graphical user
    interface.

    However if we try to programmatically download the file it cannot be done with the normal url.

    So, the current function converts the "read-only" link to a downloadable format.

    ----------
    Parameters
    ----------
    link : str
        Sharable Google Drive link.

    Returns
    -------
    out : str
        Manipulated link, that ensures a direct download with wget function.
    """

    # Get file id.
    if "id=" not in link:
        # Split link into segments (split character --> /)
        split_link = link.split("/")

        file_id = split_link[-2]
    else:
        # Split link into segments (split string --> "id=")
        split_link = link.split("id=")
        file_id = split_link[-1]

    return "https://drive.google.com/uc?export=download&id=" + file_id

# 01/10/2018 19h19m :)



import numpy
import os
import matplotlib.pyplot as plt
from sklearn.utils.testing import assert_array_almost_equal, \
assert_true, assert_less
import novainstrumentation

from novainstrumentation.waves import stdwave
from novainstrumentation.waves import waves
base_dir = os.path.dirname(novainstrumentation.__file__)

def test_stdwaves(show_graph = 0):
    fname = base_dir + '/code/data/xyzcal.txt'
    t, xcal, ycal, zcal = numpy.loadtxt(fname)

    if show_graph == 1:
        plt.plot(t, xcal, 'r-', t, ycal, 'b-', t, zcal, 'k-')
        plt.show()

    test_matrix = numpy.vstack([t, xcal, ycal, zcal]).transpose()
    reference_values = numpy.std(test_matrix,0)
    test_values = stdwave(test_matrix)

    assert_array_almost_equal(test_values, reference_values, 6 ,"\n ERROR: waves.stdwaves function not working properly")

def test_waves():
    t, s = numpy.loadtxt(base_dir + '/code/data/cleanecg.txt')
    test_matrix = numpy.array([20, 40, 87, 107, 156, 175, 225, 244, 291, 311, 355, 375, 418, 438, 482,
                501, 550, 569, 624, 644, 694, 713, 764, 784, 834, 854, 905, 925, 978])
    x = waves(s,  test_matrix, -10, 100)
    plt.plot(x[0,:])
    plt.show()

test_waves()





"""
List of functions intended to detect events in the acquired data.

Available Functions
-------------------
[Public]

detect_r_peaks
    Function that applies the Pan-Tompkins algorithm for detection of R peaks.
tachogram
    From ECG signal or from the list of samples where R peaks are located, this function generates
    a tachogram, i.e., a time-series with the evolution of RR interval (in seconds) along the
    acquisition.
detect_emg_activations
    In this function a single threshold algorithm is used for identifying the begin and end of each
    muscular activation period.
    EMG signal was simplified before the "sample by sample" threshold check, through smoothing and
    application of Teager Kaiser Energy Operator (TKEO).

Available Functions
-------------------
[Private]

<Pan-Tompkins Algorithm>
_ecg_band_pass_filter
    Step 1 of ECG simplification presented on Pan-Tompkins algorithm.
_differentiate
    Step 2 of ECG simplification presented on Pan-Tompkins algorithm.
_squaring
    Step 3 of ECG simplification presented on Pan-Tompkins algorithm.
_integration
    Step 4 of ECG simplification presented on Pan-Tompkins algorithm.
_buffer_ini
    Initialisation of the buffer that stores eight RR intervals (R peak detection algorithm consists
    in a sequential verification of each signal samples and rr_buffer will store the most recent
    eight RR intervals).
_buffer_update
    Update of buffer content (used in the R peak detection algorithm).
_detects_peaks
    Function where a first sequence of conditions, defined by Pan and Tompkins, are checked in order
    to obtain a temporary list of possible R peaks.
_checkup
    Checkup of a second sequence of conditions, defined by Pan and Tompkins, in order to our
    temporary list with possible R peaks be reduced to the set of definitive R peaks.
_acceptpeak
    In this function the Pan-Tompkins algorithm parameter SPK1 is updated.
_noisepeak
    In this function the Pan-Tompkins algorithm parameter NPK1 is updated.
</Pan-Tompkins Algorithm>

<Muscular Activation Detection Algorithm>
_thres_norm_reg
    Inside this function a relative value (percentage) of the muscular activation threshold is
    converted to an absolute format.

Observations/Comments
---------------------
None

/\
"""

from .conversion import raw_to_phy
from .visualise import plot, opensignals_kwargs, opensignals_style, opensignals_color_pallet
from .aux_functions import _butter_bandpass_filter, _moving_average

import numpy
from scipy.stats import linregress
from scipy.signal import filtfilt, butter

# Base packages used in OpenSignals Tools Notebooks for plotting data
from bokeh.plotting import figure, show
from bokeh.io import output_notebook
from bokeh.layouts import gridplot
output_notebook(hide_banner=True)


def detect_r_peaks(ecg_signal, sample_rate, time_units=False, volts=False, resolution=None,
                   device="biosignalsplux", plot_result=False):
    """
    Python implementation of R peak detection algorithm (proposed by Raja Selvaraj)

    ----------
    Parameters
    ----------
    ecg_signal : list
        List of ECG acquired samples.

    sample_rate : int
        Sampling frequency.

    time_units : boolean
        If true this function will return the R peak position in seconds.

    volts : boolean
        If true, then the conversion of raw units to mV will be done. Resolution need to be
        specified.

    resolution : int or None
        Selected resolution for data acquisition.

    device : str
        Specification of the device category.

    plot_result : boolean
        If true it will be presented a graphical representation of the R peak position in the ECG
        signal.

    Returns
    -------
    out : R peak position (ndarray), R peak amplitude (ndarray)
        R peak position (sample number or time instant in seconds) and amplitude (raw or mV).

    """

    if volts is True:
        if resolution is not None:
            # ecg_signal = ((ecg_signal / 2 ** resolution) - 0.5) * 3
            ecg_signal = raw_to_phy("ECG", device, ecg_signal, resolution, option="mV")
        else:
            raise RuntimeError("For converting raw units to mV is mandatory the specification of "
                               "acquisition resolution.")

    if time_units is True:
        time = numpy.linspace(0, len(ecg_signal) / sample_rate, len(ecg_signal))
    else:
        time = numpy.linspace(0, len(ecg_signal) - 1, len(ecg_signal))

    # Filtering Step of Pan-Tompkins Algorithm.
    filtered = _ecg_band_pass_filter(ecg_signal, sample_rate)

    # Differentiation Step of Pan-Tompkins Algorithm.
    differentiated = _differentiate(filtered)

    # Rectification Step of Pan-Tompkins Algorithm.
    squared = _squaring(differentiated)

    # Integration Step of Pan-Tompkins Algorithm.
    integrated = _integration(squared, sample_rate)

    rr_buffer, spk1, npk1, threshold = _buffer_ini(integrated, sample_rate)
    probable_peaks, possible_peaks = _detects_peaks(integrated, sample_rate)
    definitive_peaks = _checkup(probable_peaks, integrated, sample_rate, rr_buffer, spk1, npk1,
                                threshold)
    definitive_peaks = list(map(int, definitive_peaks))

    # Rephasing step.
    definitive_peaks_rephase = numpy.array(definitive_peaks) - 30 * (sample_rate / 1000)
    definitive_peaks_rephase = list(map(int, definitive_peaks_rephase))

    if time_units is True:
        peaks = numpy.array(time)[definitive_peaks_rephase]
    else:
        peaks = definitive_peaks_rephase

    amplitudes = numpy.array(ecg_signal)[definitive_peaks_rephase]

    # If plot is invoked by plot_result flag, then a graphical representation of the R peaks is
    # presented to the user.
    if plot_result is True:
        time_int = numpy.array(time[1:])
        integrated = numpy.array(integrated)

        fig = figure(x_axis_label='Time (s)', y_axis_label='Raw Data',
                     **opensignals_kwargs("figure"))
        fig.line(time_int, integrated, **opensignals_kwargs("line"))
        fig.circle(time_int[definitive_peaks], integrated[definitive_peaks], size=30,
                   color="#00893E", legend="Definitive Peaks")
        fig.circle(time_int[probable_peaks], integrated[probable_peaks], size=20, color="#009EE3",
                   legend="Probable Peaks")
        fig.circle(time_int[possible_peaks], integrated[possible_peaks], size=10, color="#302683",
                   legend="Possible Peaks")

        fig2 = figure(x_axis_label='Time (s)', y_axis_label='Raw Data',
                      **opensignals_kwargs("figure"))
        fig2.line(time, ecg_signal, **opensignals_kwargs("line"))
        fig2.circle(time[definitive_peaks_rephase],
                    numpy.array(ecg_signal)[definitive_peaks_rephase],
                    size=30, color=opensignals_color_pallet(), legend="Definitive Peaks")

        opensignals_style([fig, fig2])

        grid_plot = gridplot([[fig], [fig2]], **opensignals_kwargs("gridplot"))
        show(grid_plot)

    return peaks, amplitudes


def detect_emg_activations(emg_signal, sample_rate, smooth_level=20, threshold_level=10,
                           time_units=False, volts=False, resolution=None, device="biosignalsplux",
                           plot_result=False):
    """
    Python implementation of Burst detection algorithm using TKEO operator.

    ----------
    Parameters
    ----------
    emg_signal : list
        List of EMG acquired samples.

    sample_rate : int
        Sampling frequency.

    smooth_level : number
        Defines a percentage proportional to the smoothing level, i.e. the bigger this value is,
        then more smoothed is the signal.

    threshold_level : number
        Specification of the single threshold position, used for distinguishing between activation
        (above) and inactivation samples (below).

    time_units : boolean
        If true this function will return the Burst begin and end positions in seconds.

    volts : boolean
        If true, then the conversion of raw units to mV will be done. Resolution need to be
        specified.

    resolution : int
        Selected resolution for data acquisition.

    device : str
        Specification of the device category.

    plot_result : boolean
        If true it will be presented a graphical representation of the detected burst in the EMG
        signal.

    Returns
    -------
    out : bursts begin (ndarray), bursts end (ndarray)
        Begin and end of bursts (sample number or time instant in seconds).

    smooth_signal: list
        It is return the smoothed EMG signal (after the processing steps intended to simplify the
        signal).

    threshold_level: float
        The value of the detection threshold used to locate the begin and end of each muscular
        activation period.
    """

    if volts is True:
        if resolution is not None:
            emg_signal = raw_to_phy("EMG", device, emg_signal, resolution, option="mV")
            units = "mV"
        else:
            raise RuntimeError(
                "For converting raw units to mV is mandatory the specification of acquisition "
                "resolution.")
    else:
        units = "Input Units"

    if time_units is True:
        time_units_str = "Time (s)"
        time = numpy.linspace(0, len(emg_signal) / sample_rate, len(emg_signal))
    else:
        time = numpy.linspace(0, len(emg_signal) - 1, len(emg_signal))
        time_units_str = "Sample Number"

    # ----------------------------------- Baseline Removal -----------------------------------------
    pre_pro_signal = numpy.array(emg_signal) - numpy.average(emg_signal)

    # ------------------------------------ Signal Filtering ----------------------------------------
    low_cutoff = 10  # Hz
    high_cutoff = 300  # Hz

    # Application of the signal to the filter.
    pre_pro_signal = _butter_bandpass_filter(pre_pro_signal, low_cutoff, high_cutoff, sample_rate)

    # ------------------------------ Application of TKEO Operator ----------------------------------
    tkeo = []
    for i, signal_sample in enumerate(pre_pro_signal):
        if i in (0, len(pre_pro_signal) - 1):
            tkeo.append(signal_sample)
        else:
            tkeo.append(numpy.power(signal_sample, 2) - (pre_pro_signal[i + 1] *
                                                         pre_pro_signal[i - 1]))

    # Smoothing level - Size of sliding window used during the moving average process (a function
    # of sampling frequency)
    smoothing_level = int((smooth_level / 100) * sample_rate)

    # --------------------------------- Signal Rectification ---------------------------------------
    rect_signal = numpy.absolute(tkeo)

    # ------------------------------ First Moving Average Filter -----------------------------------
    rect_signal = _moving_average(rect_signal, sample_rate / 10)

    # -------------------------------- Second Smoothing Phase --------------------------------------
    smooth_signal = []
    for i in range(0, len(rect_signal)):
        if smoothing_level < i < len(rect_signal) - smoothing_level:
            smooth_signal.append(numpy.mean(rect_signal[i - smoothing_level:i + smoothing_level]))
        else:
            smooth_signal.append(0)

    # ----------------------------------- Threshold -----------------------------------------------
    avg_pre_pro_signal = numpy.average(pre_pro_signal)
    std_pre_pro_signal = numpy.std(pre_pro_signal)

    threshold_level = avg_pre_pro_signal + _thres_norm_reg(threshold_level, smooth_signal,
                                                           pre_pro_signal) * std_pre_pro_signal

    # Generation of a square wave reflecting the activation and inactivation periods.
    binary_signal = []
    for i in range(0, len(time)):
        if smooth_signal[i] >= threshold_level:
            binary_signal.append(1)
        else:
            binary_signal.append(0)

    # ------------------------------ Begin and End of Bursts --------------------------------------
    diff_signal = numpy.diff(binary_signal)
    act_begin = numpy.where(diff_signal == 1)[0]
    act_end = numpy.where(diff_signal == -1)[0]

    if time_units is True:
        time_begin = numpy.array(time)[act_begin]
        time_end = numpy.array(time)[act_end]
    else:
        time_begin = act_begin
        time_end = act_end

    # If plot is invoked by plot_result flag, then a graphical representation of the R peaks is
    # presented to the user.
    if plot_result is True:
        plot([list(time), list(time)], [list(emg_signal), list(numpy.array(binary_signal) *
                                                               numpy.max(emg_signal))],
             yAxisLabel=["Data Samples (" + units + ")"] * 2,
             x_axis_label=time_units_str, legend=["EMG Signal", "Activation Signal"])

    return time_begin, time_end, smooth_signal, threshold_level


# ==================================================================================================
# ================================= Private Functions ==============================================
# ==================================================================================================

# [Pan-Tompkins R peak detection algorithm]
def _ecg_band_pass_filter(data, sample_rate):
    """
    Bandpass filter with a bandpass setting of 5 to 15 Hz

    ---------
    Parameters
    ----------
    data : list
        List with the ECG signal samples.
    sample_rate : int
        Sampling rate at which the acquisition took place.

    Returns
    -------
    out : list
        Filtered signal.
    """
    nyquist_sample_rate = sample_rate / 2.
    normalized_cut_offs = [5/nyquist_sample_rate, 15/nyquist_sample_rate]
    b_coeff, a_coeff = butter(2, normalized_cut_offs, btype='bandpass')[:2]
    return filtfilt(b_coeff, a_coeff, data, padlen=150)


def _differentiate(data):
    """
    Derivative nearly linear between dc and 30 Hz

    ---------
    Parameters
    ----------
    data : list
        Data samples of the signal where the first derivative estimate is done.

    Returns
    -------
    out : list
        List with the differences between consecutive samples (the length of this list is equal to
        len(data) - 1).

    """
    return numpy.diff(data)


def _squaring(data):
    """
    Squaring data point by point. Nonlinear amplification, emphasizing the higher
    frequencies

    ---------
    Parameters
    ----------
    data : ndarry
        Array that contains signal samples. Each sample value will be squared.

    Returns
    -------
    out : ndarray
        Squared signal.

    """
    return data * data


def _integration(data, sample_rate):
    """
    Moving window integration. N is the number of samples in the width of the integration
    window

    ---------
    Parameters
    ----------
    data : ndarray
        Samples of the signal where a moving window integration will be applied.
    sample_rate : int
        Sampling rate at which the acquisition took place.

    Returns
    -------
    out : ndarray
        Integrated signal samples.
    """
    wind_size = int(0.080 * sample_rate)
    int_ecg = numpy.zeros_like(data)
    cum_sum = data.cumsum()
    int_ecg[wind_size:] = (cum_sum[wind_size:] - cum_sum[:-wind_size]) / wind_size
    int_ecg[:wind_size] = cum_sum[:wind_size] / numpy.arange(1, wind_size + 1)

    return int_ecg


def _buffer_ini(data, sample_rate):
    """
    Initializes the buffer with eight 1s intervals

    ---------
    Parameters
    ----------
    data : ndarray
        Pre-processed ECG signal samples.
    sample_rate : int
        Sampling rate at which the acquisition took place.

    Returns
    -------
    rr_buffer : list
        Data structure that stores eight samples (in the future this buffer will store the duration
        of eight RR intervals instead of the 1 second values defined in initialisation).
    spk1 : float
        Initial value of SPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named signal peak).
    npk1 : int
        Initial value of NPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named noise peak).
    threshold : float
        Initial value of the adaptive threshold level (relevant parameter for the application of
        specific criteria during the identification of R peaks).

    Sources
    -------
    https://www.robots.ox.ac.uk/~gari/teaching/cdt/A3/readings/ECG/Pan+Tompkins.pdf


    """
    rr_buffer = [1] * 8
    spk1 = max(data[sample_rate:2*sample_rate])
    npk1 = 0
    threshold = _buffer_update(npk1, spk1)

    return rr_buffer, spk1, npk1, threshold


def _buffer_update(npk1, spk1):
    """
    Computes threshold based on signal and noise values

    ---------
    Parameters
    ----------
    spk1 : float
        Actual value of SPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named signal peak).
    npk1 : int
        Actual value of NPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named noise peak).

    Returns
    -------
    out : float
        The updated threshold level.

    """
    threshold = npk1 + 0.25 * (spk1 - npk1)

    return threshold


def _detects_peaks(ecg_integrated, sample_rate):
    """
    Detects peaks from local maximum

    ---------
    Parameters
    ----------
    ecg_integrated : ndarray
        Array that contains the samples of the integrated signal.
    sample_rate : int
        Sampling rate at which the acquisition took place.

    Returns
    -------
    choosen_peaks : list
        List of local maximums that pass the first stage of conditions needed to be considered as
        a R peak.
    possible_peaks : list
        List with all the local maximums in the signal.

    """

    # Minimum RR interval = 200 ms
    min_rr = (sample_rate / 1000) * 200

    # Computes all possible peaks and their amplitudes
    possible_peaks = [i for i in range(0, len(ecg_integrated)-1)
                      if ecg_integrated[i-1] < ecg_integrated[i] and
                      ecg_integrated[i] > ecg_integrated[i+1]]

    possible_amplitudes = [ecg_integrated[k] for k in possible_peaks]
    chosen_peaks = []

    # Starts with first peak
    if not possible_peaks:
        raise Exception("No Peaks Detected.")
    peak_candidate_i = possible_peaks[0]
    peak_candidate_amp = possible_amplitudes[0]
    for peak_i, peak_amp in zip(possible_peaks, possible_amplitudes):
        if peak_i - peak_candidate_i <= min_rr and peak_amp > peak_candidate_amp:
            peak_candidate_i = peak_i
            peak_candidate_amp = peak_amp
        elif peak_i - peak_candidate_i > min_rr:
            chosen_peaks += [peak_candidate_i - 6]  # Delay of 6 samples
            peak_candidate_i = peak_i
            peak_candidate_amp = peak_amp
        else:
            pass

    return chosen_peaks, possible_peaks


def _checkup(peaks, ecg_integrated, sample_rate, rr_buffer, spk1, npk1, threshold):
    """
    Check each peak according to thresholds

    ---------
    Parameters
    ----------
    peaks : list
        List of local maximums that pass the first stage of conditions needed to be considered as
        a R peak.
    ecg_integrated : ndarray
        Array that contains the samples of the integrated signal.
    sample_rate : int
        Sampling rate at which the acquisition took place.
    rr_buffer : list
        Data structure that stores the duration of the last eight RR intervals.
    spk1 : float
        Actual value of SPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named signal peak).
    npk1 : int
        Actual value of NPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named noise peak).
    threshold : float
        Initial value of the adaptive threshold level (relevant parameter for the application of
        specific criteria during the identification of R peaks).

    Returns
    -------
    out : list
        List with the position of the peaks considered as R peak by the algorithm.

    """
    peaks_amp = [ecg_integrated[peak] for peak in peaks]
    definitive_peaks = []
    for i, peak in enumerate(peaks):
        amp = peaks_amp[i]

        # accept if larger than threshold and slope in raw signal
        # is +-30% of previous slopes
        if amp > threshold:
            definitive_peaks, spk1, rr_buffer = _acceptpeak(peak, amp, definitive_peaks, spk1,
                                                            rr_buffer)

        # accept as qrs if higher than half threshold,
        # but is 360 ms after last qrs and next peak
        # is more than 1.5 rr intervals away
        # just abandon it if there is no peak before
        # or after
        elif amp > threshold / 2 and list(definitive_peaks) and len(peaks) > i + 1:
            mean_rr = numpy.mean(rr_buffer)
            last_qrs_ms = (peak - definitive_peaks[-1]) * (1000 / sample_rate)
            last_qrs_to_next_peak = peaks[i+1] - definitive_peaks[-1]

            if last_qrs_ms > 360 and last_qrs_to_next_peak > 1.5 * mean_rr:
                definitive_peaks, spk1, rr_buffer = _acceptpeak(peak, amp, definitive_peaks, spk1,
                                                                rr_buffer)
            else:
                npk1 = _noisepeak(amp, npk1)
        # if not either of these it is noise
        else:
            npk1 = _noisepeak(amp, npk1)
        threshold = _buffer_update(npk1, spk1)

    definitive_peaks = numpy.array(definitive_peaks)

    return definitive_peaks


def _acceptpeak(peak, amp, definitive_peaks, spk1, rr_buffer):
    """
    Private function intended to insert a new RR interval in the buffer.

    ---------
    Parameters
    ----------
    peak : int
        Sample where the peak under analysis is located.
    amp : int
        Amplitude of the peak under analysis.
    definitive_peaks : list
        List with the definitive_peaks stored until the present instant.
    spk1 : float
        Actual value of SPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named signal peak).
    rr_buffer : list
        Data structure that stores the duration of the last eight RR intervals.

    Returns
    -------
    definitive_peaks_out : list
        Definitive peaks list.
    spk1 : float
        Updated value of SPK1 parameter.
    rr_buffer : list
        Buffer after appending a new RR interval and excluding the oldest one.

    """

    definitive_peaks_out = definitive_peaks
    definitive_peaks_out = numpy.append(definitive_peaks_out, peak)
    spk1 = 0.125 * amp + 0.875 * spk1  # spk1 is the running estimate of the signal peak
    if len(definitive_peaks_out) > 1:
        rr_buffer.pop(0)
        rr_buffer += [definitive_peaks_out[-1] - definitive_peaks_out[-2]]

    return numpy.array(definitive_peaks_out), spk1, rr_buffer


def _noisepeak(amp, npk1):
    """
    Private function intended to insert a new RR interval in the buffer.

    ---------
    Parameters
    ----------
    amp : int
        Amplitude of the peak under analysis.
    npk1 : int
        Actual value of NPK1 parameter defined in Pan-Tompkins real-time R peak detection algorithm
        (named noise peak).

    Returns
    -------
    npk1 : float
        Updated value of NPK1 parameter.
    """
    npk1 = 0.125 * amp + 0.875 * npk1  # npk1 is the running estimate of the noise peak

    return npk1


def tachogram(data, sample_rate, signal=False, in_seconds=False, out_seconds=False):
    """
    Function for generation of ECG Tachogram.

    ----------
    Parameters
    ----------
    data : list
        ECG signal or R peak list. When the input is a raw signal the input flag signal should be
        True.

    sample_rate : int
        Sampling frequency.

    signal : boolean
        If true, then the data argument contains the set of the ECG acquired samples.

    in_seconds : boolean
        If the R peaks list defined as the input argument "data" contains the sample numbers where
        the R peaks occur, then in_seconds needs to be False.

    out_seconds : boolean
        If True then each sample of the returned time axis is expressed in seconds.

    Returns
    -------
    out : list, list
        List of tachogram samples. List of instants where each cardiac cycle ends.

    """

    if signal is False:  # data is a list of R peaks position.
        data_copy = data
        time_axis = numpy.array(data)#.cumsum()
        if out_seconds is True and in_seconds is False:
            time_axis = time_axis / sample_rate
    else:  # data is a ECG signal.
        # Detection of R peaks.
        data_copy = detect_r_peaks(data, sample_rate, time_units=out_seconds, volts=False,
                                   resolution=None, plot_result=False)[0]
        time_axis = data_copy

    # Generation of Tachogram.
    tachogram_data = numpy.diff(time_axis)
    tachogram_time = time_axis[1:]

    return tachogram_data, tachogram_time


# [Muscular Activation Detection Algorithm]

def _thres_norm_reg(threshold_level, signal, pre_smooth_signal):
    """
    Regression function that with a percent input gives an absolute value of the threshold
    level (used in the muscular activation detection algorithm).
    Converts a relative threshold level to an absolute value.

    ---------
    Parameters
    ----------
    threshold_level : int
        Percentage value that defines the absolute threshold level relatively to the maximum value
        of signal.
    signal : list
        List of EMG smoothed signal samples.
    pre_smooth_signal : list
        Original EMG samples.

    Returns
    -------
    out : float
        Threshold level in absolute format.

     """
    avg_signal = numpy.average(pre_smooth_signal)
    std_signal = numpy.std(pre_smooth_signal)

    threshold_0_perc_level = (-avg_signal) / float(std_signal)
    threshold_100_perc_level = (numpy.max(signal) - avg_signal) / float(std_signal)

    slope, b_coeff = linregress([0, 100], [threshold_0_perc_level, threshold_100_perc_level])[:2]
    return slope * threshold_level + b_coeff

# 01/10/2018 19h19m :)




"""
Module responsible for the definition of functions that convert Raw units (available in the
acquisition files returned by OpenSignals) and sample units to physical units like mV, A, ºC,
s,..., accordingly to the sensor under analysis.

Available Functions
-------------------
[Public]

raw_to_phy
    Function that converts each sample value in raw units to a physical unit taking into account
    the respective transfer function for the sensor and device specified as an inputs.
generate_time
    Considering the acquisition sampling rate and the number of samples that compose
    the signal, this function will return a time axis in seconds.

Observations/Comments
---------------------
None

/\
"""

import numpy
from .aux_functions import _is_a_url, _generate_download_google_link
from .load import load


def raw_to_phy(sensor, device, raw_signal, resolution, option):
    """
    Function for converting raw units to physical units.

    ----------
    Parameters
    ----------
    sensor : str
        Sensor label:
        - "ECG"
        - "EMG"
        - "TEMP"
        - "BVP"
        - "SpO2.HEAD"
        - "SpO2.FING"
        - "SpO2.ARM"

    device : str
        Plux device label:
        - "bioplux"
        - "bioplux_exp"
        - "biosignalsplux"
        - "rachimeter"
        - "channeller"
        - "swifter"
        - "ddme_openbanplux"

    raw_signal : list
        Raw signal samples.

    resolution : int
        Resolution selected during acquisition.

    option : str (optional)
        Output units (only available in certain sensors):
        - "mV"
        - "V"
        - "C" (Celsius)
        - "K" (Kelvin)
        - "Ohm"
        - "A"
        - "uA"
        (When is not applicable a warning message is raised).

    Returns
    -------
    out : list
        Signal in the new scale.
    """
    raw_signal = numpy.array(raw_signal)

    # Check if resolution has the correct data format.
    if not isinstance(resolution, int):
        raise RuntimeError("The specified resolution needs to be an integer.")

    out = None
    if sensor == "TEMP":
        vcc = 3.0
        available_dev_1 = ["bioplux", "bioplux_exp", "biosignalsplux", "rachimeter", "channeller",
                           "swifter", "ddme_openbanplux"]
        available_dev_2 = ["bitalino", "bitalino_rev", "bitalino_riot"]
        if option == "Ohm":
            if device in available_dev_1:
                out = (1e4 * raw_signal) / (2**resolution - raw_signal)
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")
        elif option == "K":
            a_0 = 1.12764514e-3
            a_1 = 2.34282709e-4
            a_2 = 8.77303013e-8
            out = 1 / (a_0 + a_1 * numpy.log(raw_to_phy(sensor, device, list(raw_signal),
                                                        resolution, option="Ohm")) + a_2 *
                       ((numpy.log(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                              option="Ohm"))) ** 3))
        elif option == "C":
            if device in available_dev_1:
                out = numpy.array(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                             option="K")) - 273.15
            elif device in available_dev_2:
                out = ((raw_signal / (2 ** resolution)) * vcc - 0.5) * 100
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")
        else:
            raise RuntimeError("The selected output unit is invalid for the sensor under analysis.")

    elif sensor == "EMG":
        available_dev_1 = ["bioplux", "bioplux_exp", "biosignalsplux", "rachimeter", "channeller",
                           "swifter", "ddme_openbanplux"]
        available_dev_2 = ["bitalino"]
        available_dev_3 = ["bitalino_rev", "bitalino_riot"]
        if option == "mV":
            vcc = 3.0
            if device in available_dev_1:
                offset = 0.5
                gain = 1
            elif device in available_dev_2:
                offset = 0.5
                gain = 1.008
            elif device in available_dev_3:
                offset = 0.5
                gain = 1.009
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")
            out = (raw_signal * vcc / (2 ** resolution) - vcc * offset) / gain

        elif option == "V":
            out = numpy.array(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                         option="mV")) / 1000

    elif sensor == "ECG":
        available_dev_1 = ["bioplux", "bioplux_exp", "biosignalsplux", "rachimeter", "channeller",
                           "swifter", "ddme_openbanplux"]
        available_dev_2 = ["bitalino", "bitalino_rev", "bitalino_riot"]
        if option == "mV":
            vcc = 3.0
            if device in available_dev_1:
                offset = 0.5
                gain = 1.019
            elif device in available_dev_2:
                offset = 0.5
                gain = 1.1
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")
            out = (raw_signal * vcc / (2 ** resolution) - vcc * offset) / gain

        elif option == "V":
            out = numpy.array(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                         option="mV")) / 1000

    elif sensor == "BVP":
        available_dev_1 = ["bioplux", "bioplux_exp", "biosignalsplux", "rachimeter", "channeller",
                           "swifter", "ddme_openbanplux"]
        if option == "uA":
            vcc = 3.0
            if device in available_dev_1:
                offset = 0
                gain = 0.190060606
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")
            out = (raw_signal * vcc / (2 ** resolution) - vcc * offset) / gain

        elif option == "A":
            out = numpy.array(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                         option="uA")) * 1e-6

    elif sensor in ["SpO2.ARM", "SpO2.HEAD", "SpO2.FING"]:
        available_dev_1 = ["channeller", "biosignalsplux", "swifter"]

        scale_factor = None
        if "ARM" in sensor or "FING" in sensor:
            scale_factor = 1.2
        elif "HEAD" in sensor:
            scale_factor = 0.15

        if option == "uA":
            if device in available_dev_1:
                out = scale_factor * (raw_signal / (2 ** resolution))
            else:
                raise RuntimeError("The output specified unit does not have a defined transfer "
                                   "function for the used device.")

        elif option == "A":
            out = numpy.array(raw_to_phy(sensor, device, list(raw_signal), resolution,
                                         option="uA")) * 1e-6

    else:
        raise RuntimeError("The specified sensor is not valid or for now is not available for unit "
                           "conversion.")

    return out


def generate_time(signal, sample_rate=1000):
    """
    Function intended to generate a time axis of the input signal.

    ----------
    Parameters
    ----------
    signal : list
        List with the signal samples.

    sample_rate : int
        Sampling frequency of acquisition.

    Returns
    -------
    out : list
        Time axis with each list entry in seconds.
    """

    # Download of signal if the input is a url.
    if _is_a_url(signal):
        # Check if it is a Google Drive sharable link.
        if "drive.google" in signal:
            signal = _generate_download_google_link(signal)
        data = load(signal, remote=True)
        mac = list(data.keys())[0]
        chn = list(data[mac].keys())[0]
        signal = data[mac][chn]

    nbr_of_samples = len(signal)
    end_of_time = nbr_of_samples / sample_rate

    # ================================= Generation of the Time Axis ===============================
    time_axis = numpy.linspace(0, end_of_time, nbr_of_samples)

    return list(time_axis)

# 25/09/2018 18h58m :)



from numpy import *
import numpy as np

#testar
#escrever comentários:
    # dizer o que a função faz
    # referir valores negativos faz com que a função segmente de valores anteriores ao evento até valores superiores

# sugestões de melhoria:
    # testar se lowerBound < upperBound
    # aviso de segmentos ignorados quando [(upperBound - lowerBound) == len(i)] == False
    # bloquear o inicio e o fim da funcao

def waves(signal, events, lowerBound, upperBound):
    signal = [signal[(center + lowerBound):(center + upperBound)] for center in events]
    x = np.array(list(filter(lambda i: (upperBound - lowerBound) == len(i), signal)))
    return x


"""
File that specifies package requisites for a correct functioning of all available functions.

"""

from setuptools import setup
from os import path
from sys import platform

# Adjust requirements accordingly to the operating system.
requirements = ['numpy', 'matplotlib', 'scipy', 'h5py', 'python-magic', 'wget', 'datetime',
      'libmagic', 'bokeh', 'scipy', 'IPython', 'pandas', 'nbformat', 'ipython',
      'python-magic-bin;platform_system!="Linux"', 'requests']

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README_BSN.md'), encoding='utf-8') as f:
      long_description = f.read()


setup(name='biosignalsnotebooks',
      version='0.4.6',#major.minor.build_nbr
      description='A Python package for supporting the external loading and processing of '
                  'OpenSignals electrophysiological acquisitions.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/biosignalsnotebooks/biosignalsnotebooks',
      author='Plux Wireless Biosignals',
      author_email='gramos@plux.info',
      license='MIT',
      packages=['biosignalsnotebooks'],
      platforms=["win32", "darwin", "os2", "os2emx"],
      setup_requires=['numpy'],
      install_requires=requirements,
      zip_safe=False,
      include_package_data=True)

# 02/11/2018  17h33m :)



from numpy import *
import numpy as np
from scipy import interpolate
from pylab import *

def frequencyAlignDistance(f1,w1,f2,w2):
    
    w1=smooth(w1)
    w2=smooth(w2)
    
#    return msedistance(w2[:100],w1[:100])
    
    m1=argmax(w1)
    m2=argmax(w2)
    
    f1=f1/float(f1[m1])
    f2=f2/float(f2[m2])
    
    w1=w1/w1[m1]
    w2=w2/w2[m2]
    
#    plot(f1,w1)
#    plot(f2,w2)
    
    nf=f1[3:103]
        
    f_interp_w2=interpolate.interp1d(f2,w2)
    
    iw2=f_interp_w2(nf)
    
    plot(nf,iw2)
    plot(nf,w1[3:103])
    
    s2=iw2
    s1=w1[3:103]
    
    r=np.sqrt(np.sum(((s1-s2)**2)/s1 ))
    
    return r
    
#    return msedistance(iw2,w1[3:103])


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 20 16:20:03 2013

@author: utilizador
"""

import sys
from numpy import NaN, Inf, arange, isscalar, array, asarray
    
##############################################################################
########################### Peaks Detection ##################################
##############################################################################

def peakdelta(v, delta, x=None):
    """
    Returns two arrays

    function [maxtab, mintab]=peakdelta(v, delta, x)
    %PEAKDET Detect peaks in a vector
    %        [MAXTAB, MINTAB] = peakdelta(V, DELTA) finds the local
    %        maxima and minima ("peaks") in the vector V.
    %        MAXTAB and MINTAB consists of two columns. Column 1
    %        contains indices in V, and column 2 the found values.
    %      
    %        With [MAXTAB, MINTAB] = peakdelta(V, DELTA, X) the indices
    %        in MAXTAB and MINTAB are replaced with the corresponding
    %        X-values.
    %
    %        A point is considered a maximum peak if it has the maximal
    %        value, and was preceded (to the left) by a value lower by
    %        DELTA.
    %        Eli Billauer, 3.4.05 (Explicitly not copyrighted).

    %        This function is released to the public domain; Any use is allowed.
    """
    maxtab = []
    mintab = [] 
    
    if x is None:
        x = arange(len(v))
    
    v = asarray(v)
    
    if len(v) != len(x):
        sys.exit('Input vectors v and x must have same length')

    if not isscalar(delta):
        sys.exit('Input argument delta must be a scalar')

    if delta <= 0:
        sys.exit('Input argument delta must be positive')

    mn, mx = Inf, -Inf
    mnpos, mxpos = NaN, NaN

    lookformax = True

    for i in arange(len(v)):
        this = v[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]

        if lookformax:
            if this < mx - delta:
                maxtab.append((mxpos, mx))
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn + delta:
                mintab.append((mnpos, mn))
                mx = this
                mxpos = x[i]
                lookformax = True

    return array(maxtab), array(mintab)




"""
    File that contains the Python code to be executed in "signal_samples" Notebook.
"""


IMPORT_CODE = """
import biosignalsnotebooks as bsnb
import numpy

# Base packages used in OpenSignals Tools Notebooks for plotting data
from bokeh.plotting import figure, output_file, show
from bokeh.io import output_notebook
from bokeh.layouts import gridplot
from bokeh.models.tools import *
output_notebook(hide_banner=True)
"""

PLOTS_SIGNAL_SAMPLES = """\
signal_dict, file_header = bsnb.load(signal_samples_dir + file, get_header=True)
mac_addresses = list(signal_dict.keys())

mac_0 = mac_addresses[0]
chn_0 = list(signal_dict[mac_0].keys())[0]
sample_rate = file_header[mac_0]["sampling rate"]
time = numpy.linspace(0, len(signal_dict[mac_0][chn_0]) / sample_rate,
                      len(signal_dict[mac_0][chn_0]))
grid_layout = []
for mac in mac_addresses:
    channels = list(signal_dict[mac].keys())
    for chn in channels:
        fig = figure(x_axis_label='Time (s)', y_axis_label='Raw Data',
                     title=mac + "@" + chn, **bsnb.opensignals_kwargs("figure"))
        fig.line(time, signal_dict[mac][chn],
                 **bsnb.opensignals_kwargs("line"))
        grid_layout.append([fig])
bsnb.opensignals_style([item for sublist in grid_layout for item in sublist])
grid_plot = gridplot(grid_layout, **bsnb.opensignals_kwargs("gridplot"))
show(grid_plot)"""

# 07/11/2018  00h02m :)




def plotheatmap2(s,valmin,valmax,points=100):
    
    import numpy as np
    import matplotlib.cm as cm
    import matplotlib.mlab as mlab
    import matplotlib.pylab as plt

    x = range(1,len(s))
    y = np.linspace(valmin, valmax, points)
    X, Y = np.meshgrid(x, y)
    Z = mlab.bivariate_normal(X, Y)*0.0

    for i in range(len(s)):
        Z1 = mlab.bivariate_normal(X, Y, 3.0, 3.0, i, s[i])
        Z=Z1+Z

    im = plt.imshow(Z, interpolation='bilinear', cmap=cm.gray,
                origin='lower')
#extent=[-3,3,-3,3])
    return Z


"""
novainstrumentation

peak finding module
"""

from numpy import array, clip, argsort, sort
import numpy as np
#from pylab import find
#from scipy.signal import argrelmax

#### FROM SCIPY
#### Given that the new version failed to compile in some linux versions



def _boolrelextrema(data, comparator,
                  axis=0, order=1, mode='clip'):
    """
    Calculate the relative extrema of `data`.

    Relative extrema are calculated by finding locations where
    ``comparator(data[n], data[n+1:n+order+1])`` is True.

    Parameters
    ----------
    data : ndarray
        Array in which to find the relative extrema.
    comparator : callable
        Function to use to compare two data points.
        Should take 2 numbers as arguments.
    axis : int, optional
        Axis over which to select from `data`.  Default is 0.
    order : int, optional
        How many points on each side to use for the comparison
        to consider ``comparator(n,n+x)`` to be True.
    mode : str, optional
        How the edges of the vector are treated.  'wrap' (wrap around) or
        'clip' (treat overflow as the same as the last (or first) element).
        Default 'clip'.  See numpy.take

    Returns
    -------
    extrema : ndarray
        Indices of the extrema, as boolean array of same shape as data.
        True for an extrema, False else.

    See also
    --------
    argrelmax, argrelmin

    Examples
    --------
    array([False, False,  True, False, False], dtype=bool)

    """
    if((int(order) != order) or (order < 1)):
        raise ValueError('Order must be an int >= 1')

    datalen = data.shape[axis]
    locs = np.arange(0, datalen)

    results = np.ones(data.shape, dtype=bool)
    main = data.take(locs, axis=axis, mode=mode)
    for shift in iter(range(1, order + 1)):
        plus = data.take(locs + shift, axis=axis, mode=mode)
        minus = data.take(locs - shift, axis=axis, mode=mode)
        results &= comparator(main, plus)
        results &= comparator(main, minus)
        if(~results.any()):
            return results
    return results


def argrelmin(data, axis=0, order=1, mode='clip'):
    """
    Calculate the relative minima of `data`.

    .. versionadded:: 0.11.0

    Parameters
    ----------
    data : ndarray
        Array in which to find the relative minima.
    axis : int, optional
        Axis over which to select from `data`.  Default is 0.
    order : int, optional
        How many points on each side to use for the comparison
        to consider ``comparator(n, n+x)`` to be True.
    mode : str, optional
        How the edges of the vector are treated.
        Available options are 'wrap' (wrap around) or 'clip' (treat overflow
        as the same as the last (or first) element).
        Default 'clip'. See numpy.take

    Returns
    -------
    extrema : ndarray
        Indices of the minima, as an array of integers.

    See also
    --------
    argrelextrema, argrelmax

    Notes
    -----
    This function uses `argrelextrema` with np.less as comparator.

    """
    return argrelextrema(data, np.less, axis, order, mode)


def argrelmax(data, axis=0, order=1, mode='clip'):
    """
    Calculate the relative maxima of `data`.

    .. versionadded:: 0.11.0

    Parameters
    ----------
    data : ndarray
        Array in which to find the relative maxima.
    axis : int, optional
        Axis over which to select from `data`.  Default is 0.
    order : int, optional
        How many points on each side to use for the comparison
        to consider ``comparator(n, n+x)`` to be True.
    mode : str, optional
        How the edges of the vector are treated.
        Available options are 'wrap' (wrap around) or 'clip' (treat overflow
        as the same as the last (or first) element).
        Default 'clip'.  See `numpy.take`.

    Returns
    -------
    extrema : ndarray
        Indices of the maxima, as an array of integers.

    See also
    --------
    argrelextrema, argrelmin

    Notes
    -----
    This function uses `argrelextrema` with np.greater as comparator.

    """
    return argrelextrema(data, np.greater, axis, order, mode)


def argrelextrema(data, comparator, axis=0, order=1, mode='clip'):
    """
    Calculate the relative extrema of `data`.

    .. versionadded:: 0.11.0

    Parameters
    ----------
    data : ndarray
        Array in which to find the relative extrema.
    comparator : callable
        Function to use to compare two data points.
        Should take 2 numbers as arguments.
    axis : int, optional
        Axis over which to select from `data`.  Default is 0.
    order : int, optional
        How many points on each side to use for the comparison
        to consider ``comparator(n, n+x)`` to be True.
    mode : str, optional
        How the edges of the vector are treated.  'wrap' (wrap around) or
        'clip' (treat overflow as the same as the last (or first) element).
        Default is 'clip'.  See `numpy.take`.

    Returns
    -------
    extrema : ndarray
        Indices of the extrema, as an array of integers (same format as
        np.argmin, np.argmax).

    See also
    --------
    argrelmin, argrelmax

    """
    results = _boolrelextrema(data, comparator,
                              axis, order, mode)
    if ~results.any():
        return (np.array([]),) * 2
    else:
        return np.where(results)


#### FROM SCIPY 



def peaks(signal, tol=None):
    """ This function detects all the peaks of a signal and returns those time
    positions. To reduce the amount of peaks detected, a threshold is
    introduced so only the peaks above that value are considered.

    Parameters
    ----------
    x: array-like
     the signal with the peaks to detect.
    tol: int
      the threshold used to limit the peak detection. in case none is given,
      the value used is the minimum of the signal (detection of peaks in all
      signal)

    Returns
    -------
    peaks: array-like
      the time sample where the peak occurs.

    Example
    -------
    >>> peaks([1,2,4,3,2,5,7,7,4,9,2])
    array([2, 9])
    >>> peaks([1,2,-4,-3,-5,4,5])
    array([1, 3])
    >>> peaks([1,-4,-3,4,5],0)
    array([], dtype=int32)
    """

    if (tol is None):
        tol = min(signal)
    pks = argrelmax(clip(signal, tol, signal.max()))
    return pks[0]


#def post_peak(v, v_post):
#    """ Detects the next peak """
#    return array([v_post[find(v_post > i)[0]] for i in v])


#def prior_peak(v, v_prior):
#    """ Detects the previous peak """
#    return array([v_prior[find(v_prior < i)[-1]] for i in v])


def clean_near_peaks(signal, peaks_, min_distance):
    """ Given an array with all the peaks of the signal ('peaks') and a
    distance value ('min_distance') and the signal, by argument, this function
    erases all the unnecessary peaks and returns an array with only the maximum
    peak for each period of the signal (the period is given by the
    min_distance).

    Parameters
    ----------
    signal: array-like
      the original signal.
    peaks: array-like
      the peaks to filter.
    min_distance: int
      the distance value to exclude the unnecessary peaks.

    Returns
    -------
    fp: array-like
      the new peaks, after filtering just the maximum peak per period.

    See also: clean_near_events()
    """

    #order all peaks
    
    ars = argsort(signal[peaks_])
    
    pp = peaks(ars)

    fp = []

    #clean near peaks
    while len(pp) > 0:
        fp += [pp[-1]]
        pp = pp[abs(pp - pp[-1]) > min_distance]

    return sort(array(fp))


def clean_near_events(points, min_distance):
    """ Given an array with some specific points of the signal and a distance
    value, this function erases all the surplus points and returns an array
    with only one point (the first point) per distance samples values

    Parameters
    ----------
    points: array-like
      the events to filter.
    min_distance: int
      the distance value to exclude the unnecessary events.

    Returns
    -------
    fp: array-like
      the new events, after filtering just one per period.
    Example
    -------
    >>> clean_near_events([1,3,5,50,65,68,83,88],10)
    array([ 1, 50, 65, 83])

    See also: clean_near_peaks()
    """

    fp = []
    points = array(points)
    while len(points) > 0:
        fp += [points[0]]
        points = points[abs(points - points[0]) > min_distance]

    return array(fp)
    
from numpy import ceil

def bigPeaks(s,th,min_peak_distance=5,peak_return_percentage=0.1):
    p=peaks(s,th)
    pp=[]
    if len(p)==0:
        pp=[]
    else:
        p=clean_near_peaks(s,p,min_peak_distance)
    
        if len(p)!=0:
            ars=argsort(s[p])
            pp=p[ars]
            
            num_peaks_to_return=ceil(len(p)*peak_return_percentage)
            
            pp=pp[-int(num_peaks_to_return):]
        else:
            pp==[]
    return pp



import os
import numpy


import novainstrumentation

from novainstrumentation import load_with_cache

from sklearn.utils.testing import assert_array_almost_equal, \
assert_true, assert_less
# assert_almost_equal, assert_array_equal, assert_equal


base_dir = os.path.dirname(novainstrumentation.__file__)


def test_load_with_cache():
    fname = base_dir + '/code/data/bvp.txt'
    fname_npy = fname + '.npy'

    direct_d = numpy.loadtxt(fname)

    cache_d = load_with_cache(fname, data_type='float')

    #Check that the data was correctly read
    assert_array_almost_equal(direct_d, cache_d, decimal=5)

    #Guarantee that the file was created
    assert_true(os.path.exists(fname_npy))

    #The file should be smaller that the original
    assert_less(os.path.getsize(fname_npy), os.path.getsize(fname))

    os.remove(fname_npy)


def test_fail():
    assert False



"""
File with the set of strings that contain a markdown or code cell content for general application
in the notebooks.

"""

HEADER = """<table width="100%">
    <tr style="border-bottom:solid 2pt #009EE3">
        <td style="text-align:left" width="10%">
            <a href="FILENAME" download><img src="../../images/icons/download.png"></a>
        </td>
        <td style="text-align:left" width="10%">
            <a href="SOURCE" target="_blank"><img src="../../images/icons/program.png" title="Be creative and test your solutions !"></a>
        </td>
        <td></td>
        <td style="text-align:left" width="5%">
            <a href="../MainFiles/biosignalsnotebooks.ipynb"><img src="../../images/icons/home.png"></a>
        </td>
        <td style="text-align:left" width="5%">
            <a href="../MainFiles/contacts.ipynb"><img src="../../images/icons/contacts.png"></a>
        </td>
        <td style="text-align:left" width="5%">
            <a href="https://github.com/biosignalsnotebooks/biosignalsnotebooks" target="_blank"><img src="../../images/icons/github.png"></a>
        </td>
        <td style="border-left:solid 2pt #009EE3" width="15%">
            <img src="../../images/ost_logo.png">
        </td>
    </tr>
</table>"""

# FOOTER = """<hr>
# <table width="100%">
#     <tr>
#         <td style="border-right:solid 3px #009EE3" width="20%">
#             <img src="../../images/ost_logo.png">
#         </td>
#         <td width="40%" style="text-align:left">
#             <a href="https://github.com/biosignalsnotebooks/biosignalsnotebooks">&#9740; GitHub Repository</a>
#             <br>
#             <a href="../MainFiles/biosignalsnotebooks.ipynb">&#9740; Notebook Categories</a>
#             <br>
#             <a href="https://pypi.org/project/biosignalsnotebooks/">&#9740; How to install biosignalsnotebooks Python package ?</a>
#             <br>
#             <a href="../MainFiles/signal_samples.ipynb">&#9740; Signal Library</a>
#         </td>
#         <td width="40%" style="text-align:left">
#             <a href="../MainFiles/by_diff.ipynb">&#9740; Notebooks by Difficulty</a>
#             <br>
#             <a href="../MainFiles/by_signal_type.ipynb">&#9740; Notebooks by Signal Type</a>
#             <br>
#             <a href="../MainFiles/by_tag.ipynb">&#9740; Notebooks by Tag</a>
#             <br>
#             <br>
#         </td>
#     </tr>
# </table>"""

DESCRIPTION_SIGNAL_SAMPLES = """With *Plux* acquisition systems, a vast set of physiological signals can be acquired.

All the signals that were used in **<span style="color:#009EE3">biosignalsnotebooks</span>** notebooks have been collected with **bitalino** or **biosignalsplux**, being this page a resource where relevant characteristics of each acquisition are presented, together with a temporal segment of the signal."""

DESCRIPTION_GROUP_BY = """Each Notebook content is summarized in his header through a quantitative scale ('"Difficulty" between 1 and 5 stars) and some keywords (Group of "tags").

Grouping Notebooks by difficulty level, by signal type to which it applies or by tags is an extremelly important task, in order to ensure that the **<span style="color:#009EE3">biosignalsnotebooks</span>** user could navigate efficiently in this learning environment.
"""

HEADER_ALL_CATEGORIES = """<link rel="stylesheet" href="../../styles/theme_style.css">
<!--link rel="stylesheet" href="../../styles/header_style.css"-->
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css">

<table width="100%">
    <tr>
        <td id="image_td" width="15%" class="header_image_color_i"><div id="image_img"
        class="header_image_i"></div></td>
        <td class="header_text"> Notebook Title </td>
    </tr>
</table>"""

HEADER_MAIN_FILES = """<link rel="stylesheet" href="../../styles/theme_style.css">
<!--link rel="stylesheet" href="../../styles/header_style.css"-->
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css">

<table width="100%">
    <tr>
        <td id="image_td" width="20%" class="header_image_color_2">
            <img id="image_img" src="../../images/ost_logo.png"></td>
        <td class="header_text"> Notebook Title </td>
    </tr>
</table>"""

OPEN_IMAGE = """<link rel="stylesheet" href="../../styles/theme_style.css">
<img src="../../images/OS_logo_title_slim.png">
<div class="title"><h2 class="color11"> Available Notebooks </h2></div>"""

DESCRIPTION_CATEGORY = """<br>**<span style="color:#009EE3">biosignalsnotebooks</span>** (<a href="../MainFiles/aux_files/biosignalsnotebooks_presentation.pdf">see project presentation<img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a>) is a set of documents and a **<span class="color1">Python</span>** library to provide programming examples in the form of **<span class="color5">Jupyter Notebooks</span>**, as companion to the **<span style="color:#009EE3">OpenSignals</span>** biosignals acquisition tools.

This collection of code samples has the purpose to help users of PLUX Wireless Biosignals systems, such as **bitalino** or **biosignalsplux**, and to the researcher or student interested on recording processing and classifying biosignals. The examples are set on a level of complexity to inspire the users and programmers on how easy some tasks are and that more complex ones can also be achieved, by reusing and recreating some of the examples presented here.

A **<span class="color1">Python</span>** library (entitled **<span style="color:#009EE3">biosignalsnotebooks</span>** ) is the base toolbox to support the notebooks and to provide some useful functionalities. It can be installed through pip command, like demonstrated in a <a href="https://pypi.org/project/biosignalsnotebooks/" target="_blank">PyPI <img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a> dedicated page.

In many cases we also point and illustrate with code the usage of other python toolboxes dedicated to biosignal processing.

The notebooks will cover the full topics pipeline of working with biosignals, such as: **<span class="color1">Load</span>** a file; **<span class="color3">Visualise</span>** the data online and offline, **<span class="color4">Pre-Process</span>** a one channel signal or a multi-channel acquisition, **<span class="color5">Detect</span>** relevant events in the signals, **<span class="color6">Extract</span>** features from many different type of sensors and domains, **<span class="color7">Train and Classify</span>** among a set of classes with several machine learning approaches, **<span class="color8">Understand</span>** the obtained results with metrics and validations techniques.

These examples are carried in a multitude of biosignals , from ECG, EDA, EMG, Accelerometer, Respiration among many others.
The notebooks have a set of labels to help navigate among topics <a href="../MainFiles/by_tag.ipynb"><img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a>, types of signals <a href="../MainFiles/by_signal_type.ipynb"><img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a>, application area <a href="../MainFiles/biosignalsnotebooks.ipynb"><img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a> and complexity <a href="../MainFiles/by_diff.ipynb"><img src="../../images/icons/link.png" width="10px" height="10px" style="display:inline"></a> level to support the search for particular solutions.

We encourage you to share new example ideas, to pose questions info@plux.info, and to make improvements or suggestion to this set of notebooks.

**Be inspired on how to make the most of your biosignals!**

<br>
"""

HEADER_TAGS = """<div id="flex-container">
    <div id="diff_level" class="flex-item">
        <strong>Difficulty Level:</strong>   <span class="fa fa-star 1"></span>
                                <span class="fa fa-star 2"></span>
                                <span class="fa fa-star 3"></span>
                                <span class="fa fa-star 4"></span>
                                <span class="fa fa-star 5"></span>
    </div>
    <div id="tag" class="flex-item-tag">
        <span id="tag_list">
            <table id="tag_list_table">
                <tr>
                    <td class="shield_left">Tags</td>
                    <td class="shield_right" id="tags">tags</td>
                </tr>
            </table>
        </span>
        <!-- [OR] Visit https://img.shields.io in order to create a tag badge-->
    </div>
</div>"""

SEPARATOR = """<hr>"""

AUX_CODE_MESSAGE = """<span class="color6">**Auxiliary Code Segment (should not be replicated by
the user)**</span>"""

CSS_STYLE_CODE = """from biosignalsnotebooks.__notebook_support__ import css_style_apply
css_style_apply()"""

JS_CODE_AUTO_PLAY = """%%html
<script>
    // AUTORUN ALL CELLS ON NOTEBOOK-LOAD!
    require(
        ['base/js/namespace', 'jquery'],
        function(jupyter, $) {
            $(jupyter.events).on("kernel_ready.Kernel", function () {
                console.log("Auto-running all cells-below...");
                jupyter.actions.call('jupyter-notebook:run-all-cells-below');
                jupyter.actions.call('jupyter-notebook:save-notebook');
            });
        }
    );
</script>"""

FOOTER = """<hr>
<table width="100%">
    <tr>
        <td style="border-right:solid 3px #009EE3" width="20%">
            <img src="../../images/ost_logo.png">
        </td>
        <td width="40%" style="text-align:left">
            <a href="../MainFiles/aux_files/biosignalsnotebooks_presentation.pdf" target="_blank">&#9740; Project Presentation</a>
            <br>
            <a href="https://github.com/biosignalsnotebooks/biosignalsnotebooks" target="_blank">&#9740; GitHub Repository</a>
            <br>
            <a href="https://pypi.org/project/biosignalsnotebooks/" target="_blank">&#9740; How to install biosignalsnotebooks Python package ?</a>
            <br>
            <a href="../MainFiles/signal_samples.ipynb">&#9740; Signal Library</a>
        </td>
        <td width="40%" style="text-align:left">
            <a href="../MainFiles/biosignalsnotebooks.ipynb">&#9740; Notebook Categories</a>
            <br>
            <a href="../MainFiles/by_diff.ipynb">&#9740; Notebooks by Difficulty</a>
            <br>
            <a href="../MainFiles/by_signal_type.ipynb">&#9740; Notebooks by Signal Type</a>
            <br>
            <a href="../MainFiles/by_tag.ipynb">&#9740; Notebooks by Tag</a>
        </td>
    </tr>
</table>"""

MD_EXAMPLES = """<p class="steps">1 - Description of the first instruction</p>
<strong><span class="color1">Example of a Markdown cell (supports HTML and some LaTex syntax)</span></strong>"""

CODE_EXAMPLES = """# Code cell where Python instructions can be applied
print("Hello and Welcome to biosignalsnotebooks environment !")"""

# 07/11/2018  00h02m :)



# pylint: disable=C0103

import numpy as np

from novainstrumentation.panthomkins.butterworth_filters import butter_bandpass_filter
from novainstrumentation.panthomkins.detect_panthomkins_peaks import detect_panthomkins_peaks
from novainstrumentation.panthomkins.rr_update import rr_1_update, rr_2_update, sync


def panthomkins(ecg_signal, fs, butterlow=8, butterhigh=30):
    ecg = ecg_signal - np.mean(ecg_signal)
    ecg = ecg / max(ecg)

    N = len(ecg)

    # Band Pass Filter
    ecg_filter = butter_bandpass_filter(ecg, butterlow, butterhigh, fs)

    # Squaring
    ecg_filter = 50.0 * ecg_filter ** 2.0

    # Find Peaks
    pksInd = detect_panthomkins_peaks(ecg_filter, mpd=35)

    pks = ecg_filter[pksInd]

    SPKI = np.mean(pks) * 0.5
    NPKI = np.mean(pks) * 0.1

    threshold1 = NPKI + 0.25 * (SPKI - NPKI)
    threshold2 = 0.5 * threshold1

    # %Assuming an average of 78 BPM, then the time between points is 1.3 ->
    # %1.3*fs = number of points between ind;

    rr_1 = np.ones(8) * 1.3 * fs
    rr_average_1 = np.mean(rr_1)

    rr_2 = np.ones(8) * 1.3 * fs

    rr_average_2 = np.mean(rr_2)

    rr_low_limit = 0.92 * rr_average_2
    rr_high_limit = 1.16 * rr_average_2

    NPeaks = len(pksInd)

    Found = np.ones((NPeaks, 3))
    Found[:, 1] = 1.3 * fs

    NFound = 0
    NFound_Old = NFound - 1

    flag = 0
    back = 0
    ii = 0

    while NPeaks - ii > 0:
        ii += 1
        # if peak found and didn't came back to check the peak again, use
        # threshold 1
        if ii - back > 0:
            TH = threshold1
        # use threshold 2
        else:
            TH = threshold2

        # if threshold inferior to the peak amplitude
        if pks[ii - 1] >= TH:
            # found 1 peak
            NFound += 1
            # fill the found array with [peak index, peak amplitude, cycle
            # step]
            Found[NFound - 1, :] = np.r_[pksInd[ii - 1], pks[ii - 1], ii - 1]

            # Update threshold
            # if not needed to go back:
            if ii - back > 0:
                SPKI = 0.125 * pks[ii - 1] + 0.875 * SPKI
            else:
                SPKI = 0.25 * pks[ii - 1] + 0.75 * SPKI

        else:
            NPKI = 0.125 * pks[ii - 1] + 0.875 * NPKI

        threshold1 = NPKI + 0.25 * (SPKI - NPKI)
        threshold2 = 0.5 * threshold1

        if NFound_Old != NFound - 1:
            rr_1, rr_average_1 = rr_1_update(rr_1, NFound - 1, Found)
            rr_2, rr_average_2, flag, rr_low_limit, rr_high_limit = rr_2_update(
                rr_2, NFound - 1, Found, rr_low_limit, rr_high_limit)

            NFound_Old = NFound - 1

            if np.mod(NFound, 8) == 0:
                print(['Average of the 8 most recent HR is ',
                       str(rr_average_1 / fs * 60.0), ' (BPM)'])
                print('')

        if flag:
            print('Gap Found')

            flag = 0
            back = ii
            ii = Found[-1, 2]

    R = sync(Found, NFound, ecg, N)

    return R
